# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Wrappers for Designer into Policy.

This file should contain the core logic to run Designers
efficiently for testing, benchmarking, or production purposes. Typical
usage is via PolicySuggester.from_designer_factory.
"""

import abc
from typing import Generic, Optional, Type, TypeVar

from absl import logging
from vizier import pythia
from vizier import pyvizier as vz
from vizier._src.algorithms.core import abstractions as vza
from vizier._src.algorithms.policies import trial_caches
from vizier.interfaces import serializable


_T = TypeVar('_T')

_INCOPORATED_COMPLETED_TRIALS_IDS = 'incorporated_completed_trials_ids'


class DesignerPolicy(pythia.Policy):
  """Wraps a Designer into a pythia Policy.

  > IMPORTANT: If your Designer class is (partially) serializable, use
  > (Partially)SerializableDesignerPolicy instead.

  When a Designer cannot be (partially) serialized, we create a new Designer
  instance at the start of each `suggest()` call. The most recently used
  Designer instance can be saved in self._designer. However, it is for
  interactive analysis/debugging purposes only and never used in
  future `suggest()` calls.
  """

  def __init__(
      self,
      supporter: pythia.PolicySupporter,
      designer_factory: vza.DesignerFactory[vza.Designer],
      *,
      policy_name: str = 'DesignerPolicy',
  ):
    """Init.

    Args:
      supporter:
      designer_factory:
      policy_name: A user visible name.
    """
    self._supporter = supporter
    self._designer_factory = designer_factory
    self._policy_name = policy_name

  def suggest(self, request: pythia.SuggestRequest) -> pythia.SuggestDecision:
    logging.info(
        'Processing request for study guid %s, with max trial id %s',
        request.study_guid,
        request.max_trial_id,
    )
    designer = self._designer_factory(request.study_config)
    logging.info(
        'Study guid %s: Designer created. Getting trials from Vizier service.',
        request.study_guid,
    )
    completed = self._supporter.GetTrials(
        status_matches=vz.TrialStatus.COMPLETED
    )
    logging.info(
        'Study guid %s: got %s COMPLETED trials',
        request.study_guid,
        len(completed),
    )
    active = self._supporter.GetTrials(status_matches=vz.TrialStatus.ACTIVE)
    logging.info(
        'Study guid %s: got %s ACTIVE trials', request.study_guid, len(active)
    )
    logging.info(
        'Study guid %s: Updating designer with trials.', request.study_guid
    )
    designer.update(vza.CompletedTrials(completed), vza.ActiveTrials(active))
    self._designer = designer  # saved for debugging purposes only.
    logging.info(
        'Study guid %s: Asking designer for %s suggestions.',
        request.study_guid,
        request.count,
    )
    return pythia.SuggestDecision(
        designer.suggest(request.count), metadata=vz.MetadataDelta()
    )

  def early_stop(
      self, request: pythia.EarlyStopRequest
  ) -> pythia.EarlyStopDecisions:
    raise NotImplementedError(
        'DesignerPolicy does not support the early_stop() method.'
    )

  @property
  def name(self) -> str:
    return self._policy_name


class _SerializableDesignerPolicyBase(
    pythia.Policy, serializable.PartiallySerializable, Generic[_T], abc.ABC
):
  """Partially implemented class for wrapping a (Partially)SerializableDesigner.

  Inherited by (Partially)SerializableDesignerPolicy which is fully implemented.

  (Partially)SerializableDesignerPolicy maintains a synchronized state between
  the set of trial ids that were passed to Designer via `update()` call, and
  the Designer instance that was used for during last `suggest()` call.

  The policy's `dump()` contains the trial ids passed to update() and the
  result of `dump()` called on the wrapped Designer.

  Unlike in the basic DesignerPolicy class, this class tries to minimize
  the computation by following these steps in order:
    * Re-use the saved Designer instance from the last `suggest()` call.
    * `load()` the Designer state from the study-level metadata.
    * If either of the first two steps succeedes, then we update the Designer
    with **newly** completed and **all** active (pending) trials.
    * Otherwise, we update the Designer with all trials.

  > NOTE: This Policy itself is PartiallySerializable, i.e. it implements its
  own 'load' and 'dump' which wrap the designer_factory designer's 'load' and
  'dump' methods.
  """

  _ns_designer = 'designer'
  _ns_cache = 'cache'

  def __init__(
      self,
      problem_statement: vz.ProblemStatement,
      supporter: pythia.PolicySupporter,
      designer_factory: vza.DesignerFactory[_T],
      *,
      ns_root: str = 'designer_policy_v0',
      verbose: int = 0,
      seed: Optional[int] = None,
      policy_name: str = '_SerializableDesignerPolicyBase',
  ):
    """Init.

    Args:
      problem_statement:
      supporter:
      designer_factory: Factory function to create.
        (Partially)SerializableDesigner designers.
      ns_root: Root of the namespace where policy state is stored.
      verbose: Logging verbosity.
      seed: Random seed to be passed to the designer factory.
      policy_name: A user visible name.
    """
    self._supporter = supporter
    self._designer_factory = designer_factory
    self._ns_root = ns_root
    self._cache: trial_caches.IdDeduplicatingTrialLoader = (
        trial_caches.IdDeduplicatingTrialLoader(
            supporter, include_intermediate_measurements=False
        )
    )
    self._problem_statement = problem_statement
    self._verbose = verbose
    self._designer = None
    self._seed = seed
    self._policy_name = policy_name

  def suggest(self, request: pythia.SuggestRequest) -> pythia.SuggestDecision:
    """Perform a suggest operation.

    The order of operations is:
    1. Initialize the designer and load its state from metadata.
    2. Update the designer with newly completed trials.
    3. Generate suggestions from the designer.
    4. Dump the state of the designer and store it in metadata.

    Arguments:
      request: Pythia suggestion request objects.

    Returns:
      The suggestions from the designer.
    """
    # Note that we can avoid O(Num trials) dependency in the standard scenario,
    # by storing only the last element in a consecutive sequence, e.g.,
    # instead of storing [1,2,3,4,11,12,13,21], store: [4,13,21], but
    # we keep things simple in this pseudocode.
    self._initialize_designer(request.study_config)
    new_completed_trials = self._cache.get_newly_completed_trials(
        request.max_trial_id
    )
    active_trials = self._cache.get_active_trials()
    self.designer.update(
        completed=vza.CompletedTrials(new_completed_trials),
        all_active=vza.ActiveTrials(active_trials),
    )
    logging.info(
        (
            'Updated with %s new completed trials and %s new active trials. '
            'Trial Id cache state: %s'
        ),
        len(new_completed_trials),
        len(active_trials),
        str(self._cache),
    )
    metadata_delta = vz.MetadataDelta()
    # During the 'suggest' call the designer's state could be changed, therefore
    # the state is dumped and stored only after 'suggest' was called.
    suggestions = pythia.SuggestDecision(
        self.designer.suggest(request.count), metadata=metadata_delta
    )
    metadata_delta.on_study.ns(self._ns_root).attach(self.dump())
    return suggestions

  def early_stop(
      self, request: pythia.EarlyStopRequest
  ) -> pythia.EarlyStopDecisions:
    raise NotImplementedError(
        'PartiallySerializableDesignerPolicy does not implement early_stop().'
    )

  @property
  def designer(self) -> _T:
    if self._designer is None:
      raise ValueError(
          '`self._designer` has not been initialized!'
          'Use self._restore_designer(..) to initialize it.'
      )
    return self._designer

  @abc.abstractmethod
  def _restore_designer(self, designer_metadata: vz.Metadata) -> _T:
    """Creates a new Designer by restoring the state from `designer_metadata`.

    Args:
      designer_metadata:

    Returns:
      New Designer object.

    Raises:
      DecodeError: `designer_metadata` does not contain valid information
        to restore a Designer state.
    """

  def _initialize_designer(
      self, problem_statement: vz.ProblemStatement
  ) -> None:
    """Guarantees that `self._designer` is populated.

    This method guarantees that after a successful call, `self.designer` does
    not raise an Exception.
    This method catches all DecodeErrors.

    Args:
      problem_statement:

    Raises:
      ValueError: If problem_statement is differerent from the initially
        received problem_statement.
    """
    if self._designer is not None:
      # When the same policy object is maintained in RAM, prefer keeping
      # the designer object over restoring the state from metadata.
      # TOCONSIDER: Adding a boolean knob to turn off this behavior.
      logging.log_if(
          logging.INFO,
          (
              'Policy already has a designer. '
              'It will not attempt to load from metadata.'
          ),
          self._verbose >= 2,
      )
      return
    elif self._problem_statement != problem_statement:
      raise ValueError(
          f'{type(self)} cannot be re-used for different study configs!'
          f'Policy: {self}, previous study: {self._problem_statement} '
          f'new study: {problem_statement}'
      )

    metadata = problem_statement.metadata.ns(self._ns_root)
    try:
      self.load(metadata)
      logging.log_if(
          logging.INFO, 'Successfully decoded all states!', self._verbose >= 1
      )
    except serializable.DecodeError as e:
      logging.log_if(
          logging.INFO, 'Failed to decode state. %s', self._verbose >= 1, e
      )
      # Restart the state of the policy.
      self._designer = self._designer_factory(
          problem_statement, seed=self._seed
      )
      self._cache.clear()

  def dump(self) -> vz.Metadata:
    """Dump state.

    Returns:
      Metadata has the following namespace hierarchy:
        Namespace([self._ns_root]): contains the policy's state.
        Namespace([self._ns_root, self._ns_designer]: contains the designer's
          state.
    """
    md = vz.Metadata()
    md.ns(self._ns_designer).attach(self.designer.dump())
    md.ns(self._ns_cache).attach(self._cache.dump())
    return md

  def load(self, md: vz.Metadata) -> None:
    self._cache.load(md.ns(self._ns_cache))
    self._designer = self._restore_designer(md.ns(self._ns_designer))

  @property
  def name(self) -> str:
    return self._policy_name


# TODO: Rewrite base to separate incorporated trial logic from
# serialization logic. For now, it simply disables the serialization.
class InRamDesignerPolicy(_SerializableDesignerPolicyBase[vza.Designer]):
  """Wraps a Designer in RAM for efficient updates()."""

  # Override restore() to simply return the designer.
  def _restore_designer(self, designer_metadata: vz.Metadata) -> vza.Designer:
    if self._designer is None:
      raise ValueError(
          '`self._designer` has not been initialized!'
          ' Call self._initialize_designer(..) first.'
      )
    return self._designer

  # Override dump() since this is in ram.
  def dump(self) -> vz.Metadata:
    return vz.Metadata()


class PartiallySerializableDesignerPolicy(
    _SerializableDesignerPolicyBase[vza.PartiallySerializableDesigner]
):
  """Wraps a PartiallySerializableDesigner."""

  def _restore_designer(
      self, designer_metadata: vz.Metadata
  ) -> vza.PartiallySerializableDesigner:
    designer = self._designer_factory(self._problem_statement)
    designer.load(designer_metadata)
    return designer


class SerializableDesignerPolicy(
    _SerializableDesignerPolicyBase[vza.SerializableDesigner]
):
  """Wraps a SerializableDesigner."""

  def __init__(
      self,
      problem_statement: vz.ProblemStatement,
      supporter: pythia.PolicySupporter,
      designer_factory: vza.DesignerFactory[vza.SerializableDesigner],
      designer_cls: Type[vza.SerializableDesigner],
      *,
      ns_root: str = 'designer_policy_v0',
      verbose: int = 0,
  ):
    """Init.

    Args:
      problem_statement:
      supporter:
      designer_factory: Used when designer state cannot be restored.
      designer_cls: Type name of the designer. Its load() classmethod is called
        to restore the designer state.
      ns_root: Root of the namespace where policy state is stored.
      verbose: Logging verbosity.
    """
    super().__init__(
        problem_statement,
        supporter,
        designer_factory,
        ns_root=ns_root,
        verbose=verbose,
    )
    self._designer_cls = designer_cls

  def _restore_designer(
      self, designer_metadata: vz.Metadata
  ) -> vza.SerializableDesigner:
    return self._designer_cls.recover(designer_metadata)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for pyglove.tuner.vizier2.oss_vizier_test."""
import datetime
import os

from absl import logging
from vizier._src.pyglove import oss_vizier as vizier
from vizier._src.pyglove import vizier_test_lib
from vizier._src.service import constants
from vizier._src.service import vizier_server

from absl.testing import absltest


class OSSVizierSampleTest(vizier_test_lib.SampleTest):

  @classmethod
  def setUpClass(cls):
    super().setUpClass()
    server = vizier_server.DefaultVizierServer(
        host=os.uname()[1],
        database_url=constants.SQL_MEMORY_URL,
        early_stop_recycle_period=datetime.timedelta(seconds=0.0),
    )
    logging.info(server.endpoint)
    vizier._services.reset_for_testing()
    vizier.init(vizier_endpoint=server.endpoint)
    cls.server = server
    logging.info('Vizier service has been set up!')

  def __init__(self, *args, **kwargs):
    super().__init__(
        vizier.OSSVizierBackend,
        *args,
        builtin_multiobjective_algorithm_to_test='NSGA2',
        **kwargs,
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for convergence_curve."""

import numpy as np
from vizier import pyvizier
from vizier._src.benchmarks.analyzers import convergence_curve as convergence

from absl.testing import absltest
from absl.testing import parameterized


def _gen_trials(values):
  """Returns trials where trials[i] has empty metric name equal to values[i]."""
  trials = []
  for v in values:
    trial = pyvizier.Trial()
    trials.append(
        trial.complete(
            pyvizier.Measurement(metrics={'': pyvizier.Metric(value=v)})))
  return trials


class ConvergenceCurveTest(absltest.TestCase):

  def test_align_xs_merge_ys_on_different_lengths(self):
    c1 = convergence.ConvergenceCurve(
        xs=np.array([1, 2, 3]),
        ys=np.array([[2, 1, 1]]),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING,
    )
    c2 = convergence.ConvergenceCurve(
        xs=np.array([1]),
        ys=np.array([[3]]),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING,
    )
    aligned = convergence.ConvergenceCurve.align_xs([c1, c2])[0]

    np.testing.assert_array_equal(aligned.xs, [1, 2, 3])
    np.testing.assert_array_equal(
        aligned.ys, np.array([[2, 1, 1], [3, np.nan, np.nan]])
    )

  def test_align_xs_merge_ys_on_distinct_xvalues(self):
    c1 = convergence.ConvergenceCurve(
        xs=np.array([1, 3, 4]),
        ys=np.array([[2, 1, 1]]),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING,
    )
    c2 = convergence.ConvergenceCurve(
        xs=np.array([2]),
        ys=np.array([[3]]),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING,
    )
    aligned = convergence.ConvergenceCurve.align_xs([c1, c2])[0]

    np.testing.assert_array_equal(aligned.xs.shape, (3,))
    np.testing.assert_array_equal(
        aligned.ys, np.array([[2, 1.25, 1], [3, np.nan, np.nan]])
    )

  def test_align_xs_merge_ys_with_interpolation(self):
    c1 = convergence.ConvergenceCurve(
        xs=np.array([1, 2, 3, 4, 5]),
        ys=np.array([[2, 2, 1, 0.5, 0.5], [1, 1, 1, 1, 1]]),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING,
    )
    aligned = convergence.ConvergenceCurve.align_xs(
        [c1], interpolate_repeats=True
    )[0]

    np.testing.assert_array_equal(aligned.xs, np.array([1, 2, 3, 4, 5]))
    np.testing.assert_array_equal(
        aligned.ys, np.array([[2, 1.5, 1.0, 0.5, 0.5], [1, 1, 1, 1, 1]])
    )

  def test_extrapolate_ys_with_steps(self):
    c1 = convergence.ConvergenceCurve(
        xs=np.array([1, 2, 3, 4]),
        ys=np.array([[2, 1.5, 1, 0.5], [1, 1, 1, 1]]),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING,
    )

    extra_c1 = convergence.ConvergenceCurve.extrapolate_ys(c1, steps=2)

    np.testing.assert_array_equal(extra_c1.xs.shape, (6,))
    np.testing.assert_array_equal(
        extra_c1.ys,
        np.array([[2, 1.5, 1.0, 0.5, 0.0, -0.5], [1, 1, 1, 1, 1, 1]]),
    )

  def test_align_xs_merge_ys_on_increasing_and_dicreasing_fails(self):
    c1 = convergence.ConvergenceCurve(
        xs=np.array([1, 3, 4]),
        ys=np.array([[2, 3, 3]]),
        trend=convergence.ConvergenceCurve.YTrend.INCREASING,
    )
    c2 = convergence.ConvergenceCurve(
        xs=np.array([2]),
        ys=np.array([[3]]),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING,
    )
    with self.assertRaisesRegex(ValueError, 'increasing'):
      # pylint: disable=[expression-not-assigned]
      convergence.ConvergenceCurve.align_xs([c1, c2])[0]

  def test_align_xs_keep_ys(self):
    c1 = convergence.ConvergenceCurve(
        xs=np.array([1, 3, 4]),
        ys=np.array([[2, 1, 1], [8, 3, 1]]),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING,
    )
    c2 = convergence.ConvergenceCurve(
        xs=np.array([2]),
        ys=np.array([[3]]),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING,
    )

    aligned1, aligned2 = convergence.ConvergenceCurve.align_xs(
        [c1, c2], keep_curves_separate=True
    )

    np.testing.assert_array_equal(aligned1.xs, np.array([1.0, 2.5, 4.0]))
    np.testing.assert_array_equal(aligned2.xs, np.array([1.0, 2.5, 4.0]))

    np.testing.assert_array_equal(
        aligned1.ys, np.array([[2.0, 1.25, 1.0], [8.0, 4.25, 1.0]])
    )
    np.testing.assert_array_equal(
        aligned2.ys, np.array([[3.0, np.nan, np.nan]])
    )


class ConvergenceCurveConverterTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('maximize', pyvizier.ObjectiveMetricGoal.MAXIMIZE, [[2, 2, 3]]),
      ('minimize', pyvizier.ObjectiveMetricGoal.MINIMIZE, [[2, 1, 1]]),
  )
  def test_convert_basic(self, goal, expected):
    trials = _gen_trials([2, 1, 3])
    generator = convergence.ConvergenceCurveConverter(
        pyvizier.MetricInformation(name='', goal=goal)
    )
    curve = generator.convert(trials)
    np.testing.assert_array_equal(curve.xs, [1, 2, 3])
    np.testing.assert_array_equal(curve.ys, expected)

  @parameterized.named_parameters(
      ('maximize', pyvizier.ObjectiveMetricGoal.MAXIMIZE, [[2, 2, 3]]),
      ('minimize', pyvizier.ObjectiveMetricGoal.MINIMIZE, [[-2, -1, -1]]),
  )
  def test_convert_flip_signs(self, goal, expected):
    trials = _gen_trials([2, 1, 3])
    generator = convergence.ConvergenceCurveConverter(
        pyvizier.MetricInformation(name='', goal=goal), flip_signs_for_min=True
    )
    curve = generator.convert(trials)
    np.testing.assert_array_equal(curve.xs, [1, 2, 3])
    np.testing.assert_array_equal(curve.ys, expected)

  @parameterized.parameters(
      (
          pyvizier.ObjectiveMetricGoal.MAXIMIZE,
          2,
          [2, 1, 4, 5],
          [[2, 2, 5, 5]],
      ),
      (
          pyvizier.ObjectiveMetricGoal.MAXIMIZE,
          2,
          [4, 5, 2, 1],
          [[5, 5, 5, 5]],
      ),
      (
          pyvizier.ObjectiveMetricGoal.MAXIMIZE,
          2,
          [2, 1, 4],
          [[2, 2, 4]],
      ),
      (
          pyvizier.ObjectiveMetricGoal.MAXIMIZE,
          3,
          [2, 1, 4, 7, 9, 8, 10],
          [[4, 4, 4, 9, 9, 9, 10]],
      ),
      (
          pyvizier.ObjectiveMetricGoal.MINIMIZE,
          3,
          [3, 7, 2, 3, 0],
          [[2, 2, 2, 0, 0]],
      ),
  )
  def test_convert_with_batch_size(self, goal, batch_size, values, expected):
    trials = _gen_trials(values)
    generator = convergence.ConvergenceCurveConverter(
        pyvizier.MetricInformation(name='', goal=goal),
        flip_signs_for_min=False,
        batch_size=batch_size,
    )
    curve = generator.convert(trials)
    np.testing.assert_array_equal(curve.xs, list(range(1, len(trials) + 1)))
    np.testing.assert_array_equal(curve.ys, np.float_(expected))

  @parameterized.named_parameters(
      ('maximize', pyvizier.ObjectiveMetricGoal.MAXIMIZE, [[-np.inf, 2, 2]]),
      ('minimize', pyvizier.ObjectiveMetricGoal.MINIMIZE, [[-np.inf, -2, -1]]),
  )
  def test_convert_flip_signs_inf(self, goal, expected):
    sign = 1 if goal == pyvizier.ObjectiveMetricGoal.MINIMIZE else -1
    trials = _gen_trials([sign * np.inf, 2, 1])
    generator = convergence.ConvergenceCurveConverter(
        pyvizier.MetricInformation(name='', goal=goal), flip_signs_for_min=True
    )
    curve = generator.convert(trials)
    np.testing.assert_array_equal(curve.xs, [1, 2, 3])
    np.testing.assert_array_equal(curve.ys, expected)

  @parameterized.parameters(
      (pyvizier.ObjectiveMetricGoal.MAXIMIZE, [2, 1, 4, 5], 2),
      (pyvizier.ObjectiveMetricGoal.MAXIMIZE, [2, 1, 4], 1),
      (pyvizier.ObjectiveMetricGoal.MAXIMIZE, [2, 1, 4, 7, 9, 8, 10], 4),
      (pyvizier.ObjectiveMetricGoal.MINIMIZE, [3, 7, 2, 3, 4], 3),
  )
  def test_convert_with_state_updates(self, goal, values, split_idx):
    trials = _gen_trials(values)
    generator = convergence.ConvergenceCurveConverter(
        pyvizier.MetricInformation(name='', goal=goal),
        flip_signs_for_min=False,
    )
    curve = generator.convert(trials)

    generator = convergence.ConvergenceCurveConverter(
        pyvizier.MetricInformation(name='', goal=goal),
        flip_signs_for_min=False,
    )
    first_trials = trials[:split_idx]
    second_trials = trials[split_idx:]
    first_curve = generator.convert(first_trials)
    second_curve = generator.convert(second_trials)

    np.testing.assert_array_equal(
        curve.xs, np.concatenate((first_curve.xs, second_curve.xs), axis=-1)
    )
    np.testing.assert_array_equal(
        curve.ys, np.concatenate((first_curve.ys, second_curve.ys), axis=-1)
    )


class HypervolumeCurveConverterTest(parameterized.TestCase):

  def test_convert_with_origin_reference(self):
    generator = convergence.HypervolumeCurveConverter(
        [
            pyvizier.MetricInformation(
                name='max', goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE
            ),
            pyvizier.MetricInformation(
                name='min', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE
            ),
        ],
        reference_value=np.array([0.0]),
    )
    pytrials = []
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 4.0, 'min': 2.0})
        )
    )
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 3.0, 'min': -1.0})
        )
    )
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 4.0, 'min': -2.0})
        )
    )

    curve = generator.convert(pytrials)
    np.testing.assert_array_equal(curve.xs, [1, 2, 3])
    np.testing.assert_array_almost_equal(curve.ys, [[0.0, 3.0, 8.0]], decimal=1)

  def test_convert_with_reference(self):
    generator = convergence.HypervolumeCurveConverter(
        [
            pyvizier.MetricInformation(
                name='max', goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE
            ),
            pyvizier.MetricInformation(
                name='min', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE
            ),
        ],
        reference_value=np.array([3.0, 0.0]),
    )
    pytrials = []
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 4.0, 'min': 2.0})
        )
    )
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 3.0, 'min': -1.0})
        )
    )
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 4.0, 'min': -2.0})
        )
    )

    curve = generator.convert(pytrials)
    np.testing.assert_array_equal(curve.xs, [1, 2, 3])
    np.testing.assert_array_almost_equal(curve.ys, [[0.0, 0.0, 2.0]], decimal=1)

  def test_convert_with_none_reference(self):
    generator = convergence.HypervolumeCurveConverter([
        pyvizier.MetricInformation(
            name='max', goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE
        ),
        pyvizier.MetricInformation(
            name='min', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE
        ),
    ])
    pytrials = []
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': -np.inf, 'min': np.inf})
        )
    )
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 3.0, 'min': -1.0})
        )
    )
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 4.0, 'min': -2.0})
        )
    )

    curve = generator.convert(pytrials)
    np.testing.assert_array_equal(curve.xs, [1, 2, 3])
    np.testing.assert_array_almost_equal(curve.ys, [[0.0, 0.0, 1.0]], decimal=1)

  def test_convert_with_inf_none_reference(self):
    generator = convergence.HypervolumeCurveConverter([
        pyvizier.MetricInformation(
            name='max', goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE
        ),
        pyvizier.MetricInformation(
            name='min', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE
        ),
    ])
    pytrials = []
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': -np.inf, 'min': np.inf})
        )
    )
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': -np.inf, 'min': 2.0})
        )
    )

    curve = generator.convert(pytrials)
    np.testing.assert_array_equal(curve.xs, [1, 2])
    np.testing.assert_array_equal(curve.ys, [[0.0, 0.0]])

  def test_convert_with_state(self):
    generator = convergence.HypervolumeCurveConverter(
        [
            pyvizier.MetricInformation(
                name='max', goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE
            ),
            pyvizier.MetricInformation(
                name='min', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE
            ),
        ],
        reference_value=np.array([0.0]),
    )
    pytrials = []
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 4.0, 'min': 2.0})
        )
    )
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 5.0, 'min': -1.0})
        )
    )
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 4.0, 'min': -2.0})
        )
    )

    curve = generator.convert(pytrials)
    np.testing.assert_array_equal(curve.xs, [1, 2, 3])
    np.testing.assert_array_almost_equal(
        curve.ys, [[0.0, 5.0, 9.0]], decimal=0.5
    )

    pytrials = []
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 2.0, 'min': -0.1})
        )
    )
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 5.0, 'min': -2.0})
        )
    )

    curve = generator.convert(pytrials)
    np.testing.assert_array_equal(curve.xs, [4, 5])
    np.testing.assert_array_almost_equal(curve.ys, [[9.0, 10.0]], decimal=1)

  def test_convert_factor_with_inf(self):
    generator = convergence.HypervolumeCurveConverter(
        [
            pyvizier.MetricInformation(
                name='max', goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE
            ),
            pyvizier.MetricInformation(
                name='min', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE
            ),
        ],
        infer_origin_factor=1.0,
    )
    pytrials = []
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': -np.inf, 'min': np.inf})
        )
    )
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 3.0, 'min': -1.0})
        )
    )
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 4.0, 'min': -2.0})
        )
    )

    curve = generator.convert(pytrials)
    np.testing.assert_array_equal(curve.xs, [1, 2, 3])
    np.testing.assert_array_almost_equal(curve.ys, [[0.0, 1.0, 4.0]], decimal=1)


class MultiMetricCurveConverterTest(parameterized.TestCase):

  def test_convert_single_objective(self):
    generator = convergence.MultiMetricCurveConverter.from_metrics_config(
        metrics_config=pyvizier.MetricsConfig([
            pyvizier.MetricInformation(
                name='min', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE
            ),
            pyvizier.MetricInformation(
                name='safe',
                goal=pyvizier.ObjectiveMetricGoal.MINIMIZE,
                safety_threshold=0.1,
            ),
        ]),
        flip_signs_for_min=True,
    )
    pytrials = []
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'min': 4.0, 'safe': -2.0})
        )
    )
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'min': 3.0, 'safe': 1.0})
        )
    )
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'min': 2.0, 'safe': -1.0})
        )
    )

    curve = generator.convert(pytrials)
    np.testing.assert_array_equal(curve.xs, [1, 2, 3])
    np.testing.assert_array_almost_equal(curve.ys, [[-4.0, -4.0, -2.0]])

  def test_convert_multiobjective(self):
    generator = convergence.MultiMetricCurveConverter.from_metrics_config(
        metrics_config=pyvizier.MetricsConfig([
            pyvizier.MetricInformation(
                name='max', goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE
            ),
            pyvizier.MetricInformation(
                name='min', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE
            ),
            pyvizier.MetricInformation(
                name='safe',
                goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE,
                safety_threshold=-0.2,
            ),
        ]),
        reference_value=np.array([0.0]),
    )
    pytrials = []
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 4.0, 'min': -1.0, 'safe': 1.0})
        )
    )
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(
                metrics={'max': 5.0, 'min': -1.0, 'safe': -1.0}
            )
        )
    )
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 4.0, 'min': -2.0, 'safe': 2.1})
        )
    )

    curve = generator.convert(pytrials)
    np.testing.assert_array_equal(curve.xs, [1, 2, 3])
    np.testing.assert_array_almost_equal(curve.ys, [[4.0, 4.0, 8.0]], decimal=1)


class RestartingCurveConverterTest(absltest.TestCase):

  def test_convert_with_restart(self):
    def converter_factory():
      return convergence.HypervolumeCurveConverter(
          [
              pyvizier.MetricInformation(
                  name='max', goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE
              ),
              pyvizier.MetricInformation(
                  name='min', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE
              ),
          ],
      )

    restart_converter = convergence.RestartingCurveConverter(
        converter_factory, restart_min_trials=2, restart_rate=1.001
    )
    pytrials = []
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 0.0, 'min': 0.0})
        )
    )
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 5.0, 'min': -1.0})
        )
    )
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 4.0, 'min': -2.0})
        )
    )

    curve = restart_converter.convert(pytrials)
    np.testing.assert_array_equal(curve.xs, [1, 2, 3])
    np.testing.assert_array_almost_equal(curve.ys, [[0.0, 5.0, 9.0]], decimal=1)

    pytrials = []
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': -1.0, 'min': 1.0})
        )
    )
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 5.0, 'min': -2.0})
        )
    )

    restart_converter.convert([pytrials[0]])
    curve = restart_converter.convert([pytrials[1]])
    np.testing.assert_array_equal(curve.xs, [5])
    np.testing.assert_array_almost_equal(curve.ys, [[18.0]], decimal=0.5)

  def test_convert_no_restart(self):
    def converter_factory():
      return convergence.HypervolumeCurveConverter(
          [
              pyvizier.MetricInformation(
                  name='max', goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE
              ),
              pyvizier.MetricInformation(
                  name='min', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE
              ),
          ],
      )

    converter = convergence.RestartingCurveConverter(
        converter_factory, restart_min_trials=10
    )
    pytrials = []
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 0.0, 'min': 0.0})
        )
    )
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 5.0, 'min': -1.0})
        )
    )
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 4.0, 'min': -2.0})
        )
    )

    curve = converter.convert(pytrials)
    np.testing.assert_array_equal(curve.xs, [1, 2, 3])
    np.testing.assert_array_almost_equal(curve.ys, [[0.0, 5.0, 9.0]], decimal=1)

    pytrials = []
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': -1.0, 'min': 1.0})
        )
    )
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 5.0, 'min': -2.0})
        )
    )

    curve = converter.convert(pytrials)
    np.testing.assert_array_equal(curve.xs, [4, 5])
    np.testing.assert_array_almost_equal(curve.ys, [[9.0, 10.0]], decimal=0.5)


class WinRateComparatorTest(absltest.TestCase):

  def setUp(self):
    super(WinRateComparatorTest, self).setUp()
    xs = np.array(range(0, 20))
    xs_t = xs.reshape(1, len(xs))
    self._baseline_curve = convergence.ConvergenceCurve(
        xs=xs,
        ys=np.exp(np.array([-0.9, -1.0, -1.1]).reshape(3, 1) * xs_t),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING,
    )

    self._worse_curves = convergence.ConvergenceCurve(
        xs=xs,
        ys=np.exp(-0.5 * xs_t),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING,
    )
    self._better_curves = convergence.ConvergenceCurve(
        xs=xs,
        ys=np.exp(np.array([-1.5, -1.8, -2.0]).reshape(3, 1) * xs_t),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING,
    )

  def test_higher_quantile_curve(self):
    baseline_length = len(self._baseline_curve.xs)
    median_score = convergence.WinRateComparator(
        baseline_curve=self._baseline_curve, compared_curve=self._better_curves
    ).curve()
    higher_quantile_score = convergence.WinRateComparator(
        baseline_curve=self._baseline_curve,
        compared_curve=self._better_curves,
        compared_quantile=0.9,
    ).curve()

    self.assertEqual(median_score.ys.shape, (1, baseline_length))
    self.assertEqual(higher_quantile_score.ys.shape, (1, baseline_length))
    # Better curves should have positive efficiency.
    self.assertTrue((median_score.ys >= 0.0).all())
    # Higher quantile means better efficiency which means more positive scores.
    self.assertTrue((higher_quantile_score.ys >= median_score.ys).all())

  def test_get_winrate_score(self):
    # Higher compared quantile should increase score. Higher baseline
    # quantile should decrease score.
    median_score = convergence.WinRateComparator(
        baseline_curve=self._baseline_curve, compared_curve=self._better_curves
    ).score()
    self.assertGreaterEqual(
        convergence.WinRateComparator(
            baseline_curve=self._baseline_curve,
            compared_curve=self._better_curves,
            compared_quantile=0.9,
        ).score(),
        median_score,
    )
    self.assertLessEqual(
        convergence.WinRateComparator(
            baseline_curve=self._baseline_curve,
            compared_curve=self._better_curves,
            baseline_quantile=0.9,
        ).score(),
        median_score,
    )


class LogEfficiencyConvergenceComparatorTest(absltest.TestCase):

  def setUp(self):
    super(LogEfficiencyConvergenceComparatorTest, self).setUp()
    xs = np.array(range(0, 20))
    xs_t = xs.reshape(1, len(xs))
    self._baseline_curve = convergence.ConvergenceCurve(
        xs=xs,
        ys=np.exp(np.array([-0.9, -1.0, -1.1]).reshape(3, 1) * xs_t),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING,
    )

    self._worse_curves = convergence.ConvergenceCurve(
        xs=xs,
        ys=np.exp(-0.5 * xs_t),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING,
    )
    self._better_curves = convergence.ConvergenceCurve(
        xs=xs,
        ys=np.exp(np.array([-1.5, -1.8, -2.0]).reshape(3, 1) * xs_t),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING,
    )

  def test_get_relative_efficiency_curve(self):
    baseline_length = len(self._baseline_curve.xs)
    rel_effiency = convergence.LogEfficiencyConvergenceCurveComparator(
        baseline_curve=self._baseline_curve, compared_curve=self._better_curves
    ).curve()
    higher_quantile = convergence.LogEfficiencyConvergenceCurveComparator(
        baseline_curve=self._baseline_curve,
        compared_curve=self._better_curves,
        compared_quantile=0.9,
    ).curve()

    self.assertEqual(rel_effiency.ys.shape, (1, baseline_length))
    self.assertEqual(higher_quantile.ys.shape, (1, baseline_length))
    # Better curves should have positive efficiency.
    self.assertTrue((rel_effiency.ys >= 0.0).all())
    # Higher quantile means better efficiency which means more positive scores.
    self.assertTrue((higher_quantile.ys >= rel_effiency.ys).all())

  def test_get_relative_efficiency_flat(self):
    flat_curve = convergence.ConvergenceCurve(
        xs=np.array(range(0, 20)),
        ys=np.array([4.0, 3.0, 2.0] + [1.5] * 17).reshape(1, 20),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING,
    )
    self_eff = convergence.LogEfficiencyConvergenceCurveComparator(
        baseline_curve=flat_curve, compared_curve=flat_curve
    ).curve()
    # Relative efficiency of a curve on itself is close to 0.
    self.assertAlmostEqual(np.linalg.norm(self_eff.ys), 0.0, delta=0.1)

  def test_get_relative_efficiency_short_curve(self):
    baseline_length = len(self._baseline_curve.xs)
    short_length = round(baseline_length / 2)
    short_curve = convergence.ConvergenceCurve(
        xs=self._baseline_curve.xs[:short_length],
        ys=self._baseline_curve.ys[:, :short_length],
        trend=convergence.ConvergenceCurve.YTrend.DECREASING,
    )
    max_score = 10.3
    short_efficiency = convergence.LogEfficiencyConvergenceCurveComparator(
        baseline_curve=self._baseline_curve,
        compared_curve=short_curve,
        max_score=max_score,
    ).curve()
    self.assertEqual(short_efficiency.ys.shape, (1, baseline_length))
    self.assertEqual(float(short_efficiency.ys[:, -1]), -max_score)

  def test_get_efficiency_score(self):
    # Higher compared quantile should increase score. Higher baseline
    # quantile should decrease score.
    median_score = convergence.LogEfficiencyConvergenceCurveComparator(
        baseline_curve=self._baseline_curve, compared_curve=self._better_curves
    ).score()
    self.assertGreater(
        convergence.LogEfficiencyConvergenceCurveComparator(
            baseline_curve=self._baseline_curve,
            compared_curve=self._better_curves,
            compared_quantile=0.9,
        ).score(),
        median_score,
    )
    self.assertLess(
        convergence.LogEfficiencyConvergenceCurveComparator(
            baseline_curve=self._baseline_curve,
            compared_curve=self._better_curves,
            baseline_quantile=0.9,
        ).score(),
        median_score,
    )

  def test_effiency_score_symmetry(self):
    base_score = convergence.LogEfficiencyConvergenceCurveComparator(
        baseline_curve=self._baseline_curve, compared_curve=self._better_curves
    ).score()
    reversed_score = convergence.LogEfficiencyConvergenceCurveComparator(
        baseline_curve=self._better_curves, compared_curve=self._baseline_curve
    ).score()
    self.assertAlmostEqual(base_score, -reversed_score, delta=0.01)

  def test_efficiency_score_value(self):
    xs = self._baseline_curve.xs
    xs_t = xs.reshape(1, len(xs))

    worse_curves = convergence.ConvergenceCurve(
        xs=xs,
        ys=np.exp(-0.5 * xs_t),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING,
    )
    better_curves = convergence.ConvergenceCurve(
        xs=xs,
        ys=np.exp(np.array([-1.5, -1.8, -2.0]).reshape(3, 1) * xs_t),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING,
    )
    # Efficiency score for exponential curves can be approximated.

    self.assertGreater(
        convergence.LogEfficiencyConvergenceCurveComparator(
            baseline_curve=self._baseline_curve, compared_curve=better_curves
        ).score(),
        0.4,
    )
    self.assertLess(
        convergence.LogEfficiencyConvergenceCurveComparator(
            baseline_curve=self._baseline_curve, compared_curve=worse_curves
        ).score(),
        -0.4,
    )

  def test_comparator_failure(self):
    unknown_curve = convergence.ConvergenceCurve(
        xs=self._baseline_curve.xs, ys=self._baseline_curve.ys
    )
    with self.assertRaisesRegex(ValueError, 'increasing or decreasing'):
      convergence.LogEfficiencyConvergenceCurveComparator(
          baseline_curve=unknown_curve, compared_curve=self._baseline_curve
      )


class StandardizedWinRateConvergenceComparatorTest(parameterized.TestCase):

  @parameterized.parameters(
      {
          'ys1': np.array([[11, 12, 20, 50, 100, 300]]),
          'ys2': np.array([[1, 2, 10]]),
          'cutoff': None,
          'res': 0.0,
      },
      {
          'ys1': np.array([[1, 2, 10]]),
          'ys2': np.array([[11, 12, 20, 50, 100, 300]]),
          'cutoff': None,
          'res': 1.0,
      },
      {
          'ys1': np.array([[1, 4, 8, 10]]),
          'ys2': np.array([[2, 5, 6, 8]]),
          'cutoff': None,
          'res': 0.5,
      },
      {
          'ys1': np.array([[1, 4, 8, 10, 12]]),
          'ys2': np.array([[2, 5, 6, 8, 10]]),
          'cutoff': 1,
          'res': 0.25,
      },
      {
          'ys1': np.array([[1, 4, 8, 10, 12]]),
          'ys2': np.array([[2, 5, 6, 8, 10]]),
          'cutoff': 3,
          'res': 0.0,
      },
      {
          'ys1': np.array([[2, 5, 6, 8, 10]]),
          'ys2': np.array([[1, 4, 8, 10, 12]]),
          'cutoff': 3,
          'res': 1.0,
      },
  )
  def test_score_one_curve_above_other(self, ys1, ys2, res, cutoff):
    xs1 = np.arange(ys1.shape[1])
    xs2 = np.arange(ys2.shape[1])
    curve1 = convergence.ConvergenceCurve(
        xs=xs1, ys=ys1, trend=convergence.ConvergenceCurve.YTrend.INCREASING
    )
    curve2 = convergence.ConvergenceCurve(
        xs=xs2, ys=ys2, trend=convergence.ConvergenceCurve.YTrend.INCREASING
    )
    comparator = convergence.StandardizedWinRateConvergenceCurveComparator(
        curve1, curve2, xs_cutoff=cutoff
    )
    self.assertEqual(comparator.score(), res)


class PercentageBetterConvergenceComparatorTest(parameterized.TestCase):

  @parameterized.parameters(
      {
          'ys1': np.array([[1, 2, 3, 4], [1, 2, 3, 4]]),
          'ys2': np.array([[-1, 20, 30, 70]]),
          'res': 0.5,
      },
      {
          'ys1': np.array([[1, 2, 3, 4]]),
          'ys2': np.array([[10, 20, 30, 70]]),
          'res': 1.0,
      },
      {
          'ys1': np.array([[10, 20, 30, 70]]),
          'ys2': np.array([[1, 2, 3, 4]]),
          'res': -1.0,
      },
  )
  def test_score(self, ys1, ys2, res):
    xs1 = np.arange(ys1.shape[1])
    xs2 = np.arange(ys2.shape[1])
    curve1 = convergence.ConvergenceCurve(
        xs=xs1, ys=ys1, trend=convergence.ConvergenceCurve.YTrend.INCREASING
    )
    curve2 = convergence.ConvergenceCurve(
        xs=xs2, ys=ys2, trend=convergence.ConvergenceCurve.YTrend.INCREASING
    )
    comparator = convergence.PercentageBetterConvergenceCurveComparator(
        curve1, curve2
    )
    self.assertEqual(comparator.score(), res)


class NormalizedSimpleRegretConvergenceComparatorTest(parameterized.TestCase):

  @parameterized.parameters(
      {
          'ys1': np.array([[1, 4, 8, 10, 12]]),
          'ys2': np.array([[2, 5, 6, 8, 10]]),
          'trend': convergence.ConvergenceCurve.YTrend.INCREASING,
          'res': (10 - 12) / 12,
      },
      {
          'ys1': np.array([[1, 4, 8, 10, 12]]),
          'ys2': np.array([[2, 5, 6, 8, 1000]]),
          'trend': convergence.ConvergenceCurve.YTrend.INCREASING,
          'res': 1.0,
      },
      {
          'ys1': np.array([[1, 4, 8, 10, 1000]]),
          'ys2': np.array([[2, 5, 6, 8, 10]]),
          'trend': convergence.ConvergenceCurve.YTrend.INCREASING,
          'res': -0.5,
      },
      {
          'ys1': np.array([[11, 5, 3, 1]]),
          'ys2': np.array([[130, 4, 2, 0.5]]),
          'trend': convergence.ConvergenceCurve.YTrend.DECREASING,
          'res': (-0.5 - (-1)) / 1,
      },
  )
  def test_score(self, ys1, ys2, trend, res):
    xs1 = np.arange(ys1.shape[1])
    xs2 = np.arange(ys2.shape[1])
    curve1 = convergence.ConvergenceCurve(xs=xs1, ys=ys1, trend=trend)
    curve2 = convergence.ConvergenceCurve(xs=xs2, ys=ys2, trend=trend)
    comparator = convergence.NormalizedSimpleRegretComparator(curve1, curve2)
    self.assertAlmostEqual(comparator.score(), res, delta=0.0001)


class WinRateSimpleRegretConvergenceComparatorTest(parameterized.TestCase):

  @parameterized.parameters(
      {
          'ys1': np.array([[1, 4, 8, 10, 12]]),
          'ys2': np.array([[2, 5, 6, 8, 10]]),
          'trend': convergence.ConvergenceCurve.YTrend.INCREASING,
          'res': 0.0,
      },
      {
          'ys1': np.array([[1, 4, 8, 10, 12]]),
          'ys2': np.array([[2, 5, 6, 8, 1000]]),
          'trend': convergence.ConvergenceCurve.YTrend.INCREASING,
          'res': 1.0,
      },
      {
          'ys1': np.array([[1, 4, 8, 10, 1000]]),
          'ys2': np.array([[2, 5, 6, 8, 10]]),
          'trend': convergence.ConvergenceCurve.YTrend.INCREASING,
          'res': 0.0,
      },
      {
          'ys1': np.array([[11, 5, 3, 1]]),
          'ys2': np.array([[130, 4, 2, 0.5]]),
          'trend': convergence.ConvergenceCurve.YTrend.DECREASING,
          'res': 1.0,
      },
  )
  def test_score(self, ys1, ys2, trend, res):
    xs1 = np.arange(ys1.shape[1])
    xs2 = np.arange(ys2.shape[1])
    curve1 = convergence.ConvergenceCurve(xs=xs1, ys=ys1, trend=trend)
    curve2 = convergence.ConvergenceCurve(xs=xs2, ys=ys2, trend=trend)
    comparator = convergence.WinRateSimpleRegretComparator(curve1, curve2)
    self.assertEqual(comparator.score(), res)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for hartmann."""

import numpy as np
from vizier import pyvizier as vz
from vizier._src.benchmarks.experimenters.synthetic import hartmann
from vizier._src.benchmarks.testing import experimenter_testing
from absl.testing import absltest


class Hartmann6DTest(absltest.TestCase):

  def test_numpy_fn(self):
    np.testing.assert_allclose(
        hartmann._hartmann6d(
            np.asarray(
                [0.20169, 0.150011, 0.476874, 0.275332, 0.311652, 0.6573]
            )
        ),
        -3.32237,
        atol=1e-5,
    )

  def test_experimenter_argmin(self):
    trial = vz.Trial(
        parameters={
            f'x{i+1}': x
            for i, x in enumerate(
                [0.20169, 0.150011, 0.476874, 0.275332, 0.311652, 0.6573]
            )
        }
    )
    hartmann.Hartmann6D().evaluate([trial])
    self.assertAlmostEqual(
        trial.final_measurement_or_die.metrics.get_value('value', np.nan),
        -3.32237,
        places=5,
    )

  def test_experimenter(self):
    experimenter_testing.assert_evaluates_random_suggestions(
        self, hartmann.Hartmann6D()
    )


class Hartmann3DTest(absltest.TestCase):

  def test_numpy_fn(self):
    np.testing.assert_allclose(
        hartmann._hartmann3d(np.asarray([0.114614, 0.555649, 0.852547])),
        -3.86278,
        atol=1e-5,
    )

  def test_experimenter_argmin(self):
    trial = vz.Trial(
        parameters={
            f'x{i+1}': x for i, x in enumerate([0.114614, 0.555649, 0.852547])
        }
    )
    hartmann.Hartmann3D().evaluate([trial])
    self.assertAlmostEqual(
        trial.final_measurement_or_die.metrics.get_value('value', np.nan),
        -3.86278,
        places=5,
    )

  def test_experimenter(self):
    experimenter_testing.assert_evaluates_random_suggestions(
        self, hartmann.Hartmann3D()
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Abstractions."""

import abc
from typing import Optional, Protocol, Sequence, TypeVar

import attr
import chex
import jax
from vizier import pyvizier as vz
from vizier.interfaces import serializable

_T = TypeVar('_T')


@attr.define(frozen=True)
class CompletedTrials:
  """A group of completed trials.

  Attributes:
    trials: Completed Trials.
  """

  def __attrs_post_init__(self):
    for trial in self.trials:
      if trial.status != vz.TrialStatus.COMPLETED:
        raise ValueError(f'All trials must be completed. Bad trial:\n{trial}')

  trials: Sequence[vz.Trial] = attr.field(
      converter=tuple,
      validator=attr.validators.deep_iterable(
          attr.validators.instance_of(vz.Trial)
      ),
  )


@attr.define(frozen=True)
class ActiveTrials:
  """A group of active (a.k.a pending) trials.

  Attributes:
    trials: Active Trials.
  """

  def __attrs_post_init__(self):
    for trial in self.trials:
      if trial.status != vz.TrialStatus.ACTIVE:
        raise ValueError(f'All trials must be active. Bad trial:\n{trial}')

  trials: Sequence[vz.Trial] = attr.field(
      converter=tuple,
      default=attr.Factory(list),
      validator=attr.validators.deep_iterable(
          attr.validators.instance_of(vz.Trial)
      ),
  )


class _SuggestionAlgorithm(abc.ABC):
  """Suggestion algorithm."""

  @abc.abstractmethod
  def suggest(
      self, count: Optional[int] = None
  ) -> Sequence[vz.TrialSuggestion]:
    """Make new suggestions.

    Args:
      count: Makes best effort to generate this many suggestions. If None,
        suggests as many as the algorithm wants.

    Returns:
      New suggestions.
    """


class Designer(_SuggestionAlgorithm):
  """Interface for sequential suggestion algorithms.

  A Designer can be wrapped into a pythia `Policy` via `DesignerPolicy`.
  Prefer implementing a `Designer` interface over `Policy` interface, for
  shared error handling, performance monitoring, logging, etc.


  A big limitation of vanilla `DesignerPolicy` is that it does not retain states
  between consecutive suggestion requests. It creates a new `Designer` from
  scratch and calls `Designer.update` with all trials of the study.
  Many algorithms with a compact state, such as evolutionary search algorithms,
  can gain a huge performance boost from incremental updates. (Side note:
  GP-UCB does not fit into this category because its state includes
  all previous observations, which is not compact).

  If your Designer can take advantage of a persistent state, implement
  `(Partially)SerializableDesigner` interface, which can be wrapped into
  `(Partially)SerializableDesignerPolicy`.
  These Policies serialize the Designer's state, store it in Vizier DB, and
  load it for the next suggest operation. `Designer.update()` is called only
  with the newly completed trials (delta).

  IMPORTANT: `Designer` should not change its state inside `suggest()` (e.g. to
  incorporate its own suggestions before completion). If it does, use
  (Partially)SerializableDesigner interface.

  NOTE: When run inside a service binary, a `Designer` instance does not
  persist during the lifetime of a `Study`. This goes true even for the
  serializable variants; the states are recovered into a new `Designer`
  instance.

  NOTE: `Designer`s are designed to be used directly in benchmarks without
  a `Policy` wrapper. Create a single `Designer` instance for the entire study,
  and incrementally update its state with delta only.
  """

  @abc.abstractmethod
  def update(
      self, completed: CompletedTrials, all_active: ActiveTrials
  ) -> None:
    """Incorporates trials into the designer's state.

    Example:
      [t1, t2] # CompletedTrials
      [t3, t4] # Active Trials
      designer.update([t1], [t3])  # state includes: t1 and t3.
      designer.update([t2])        # state includes: t1 and t2 (not t3).
      designer.update([], [t3,t4]) # state includes: t1, t2, t3, and t4.
      designer.update([], [t3])    # state includes: t1, t2, and t3.

    Arguments:
      completed: COMPLETED trials that this Designer should additionaly
        incorporate.
      all_active: All ACTIVE (aka PENDING) trials in the study from its
        beginning.
    """


@attr.define(frozen=True)
class Prediction:
  """Container to hold predictions.

  The shape of the 'mean' and 'stddev' depends on the number of predictions
  and the number of objectives/metrics.

  In the single-objective case the shape is (num_predictions,).
  In the mulit-objective case the shape is (num_predictions, num_metrics).

  The metadata could be used to supply additional information about the
  prediction.
  """

  mean: chex.Array
  stddev: chex.Array
  metadata: Optional[vz.Metadata] = None

  def __attrs_post_init__(self):
    if self.mean.shape != self.stddev.shape:
      raise ValueError('The shape of mean and stddev needs to be the same.')


class Predictor(abc.ABC):
  """Predicts objective values, given suggestions.

  For algorithms which involve the use of function regressors, this class also
  acts as a mixin to expose their prediction API.
  """

  @abc.abstractmethod
  def predict(
      self,
      trials: Sequence[vz.TrialSuggestion],
      rng: Optional[jax.Array] = None,
      num_samples: Optional[int] = None,
  ) -> Prediction:
    """Returns the mean and stddev for any given suggestions.

    Arguments:
      trials: The suggestions where the predictions will be made. Can also be
        completed trials for retrospective predictions.
      rng: The sampling random key used for approximation (if applicable).
      num_samples: The number of samples used for the approximation (if
        applicable).

    Returns:
      The predictions for the given suggestions.
    """


class DesignerFactory(Protocol[_T]):
  """Protocol (PEP-544) for a designer factory."""

  def __call__(self, problem: vz.ProblemStatement, **kwargs) -> _T:
    pass


class PartiallySerializableDesigner(
    Designer, serializable.PartiallySerializable
):
  pass


class SerializableDesigner(Designer, serializable.Serializable):
  pass



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""These are the externally-useful symbols for OSS Pythia."""
# pylint: disable=unused-import

# The API for a Pythia Policy -- i.e. the algorithm that Pythia serves.
from vizier._src.pythia.local_policy_supporters import InRamPolicySupporter
from vizier._src.pythia.policy import EarlyStopDecision
from vizier._src.pythia.policy import EarlyStopDecisions
from vizier._src.pythia.policy import EarlyStopRequest
from vizier._src.pythia.policy import Policy
from vizier._src.pythia.policy import SuggestDecision
from vizier._src.pythia.policy import SuggestRequest
from vizier._src.pythia.policy_factory import PolicyFactory
from vizier._src.pythia.policy_supporter import PolicySupporter
from vizier._src.pythia.pythia_errors import CancelComputeError
from vizier._src.pythia.pythia_errors import InactivateStudyError
from vizier._src.pythia.pythia_errors import LoadTooLargeError
from vizier._src.pythia.pythia_errors import PythiaError
from vizier._src.pythia.pythia_errors import PythiaFallbackError
from vizier._src.pythia.pythia_errors import PythiaProtocolError
from vizier._src.pythia.pythia_errors import TemporaryPythiaError
from vizier._src.pythia.pythia_errors import VizierDatabaseError

# Documentation
assert issubclass(CancelComputeError, Exception)
assert issubclass(PythiaProtocolError, Exception)
assert issubclass(VizierDatabaseError, Exception)
assert issubclass(InactivateStudyError, PythiaError)
assert issubclass(LoadTooLargeError, PythiaError)
assert issubclass(TemporaryPythiaError, PythiaError)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Eagle strategy designer serialization."""

import json
from typing import Any
import attr
import numpy as np
from vizier import pyvizier as vz
from vizier._src.algorithms.designers.eagle_strategy import eagle_strategy_utils
from vizier.interfaces import serializable

FireflyPool = eagle_strategy_utils.FireflyPool
Firefly = eagle_strategy_utils.Firefly
EagleStrategyUtils = eagle_strategy_utils.EagleStrategyUtils
OBJECTIVE_NAME = eagle_strategy_utils.OBJECTIVE_NAME


class PartialFireflyPoolEncoder(json.JSONEncoder):
  """Eagle strategy pool partial encoder.

  The encoder encodes the '_pool' dictionary, 'capacity', '_last_id' and
  '_max_fly_id' into a string format.

  The encoder does not store the EagleStrategyUtils as its state only depends
  on the random generator which is handled separately.
  """

  def default(self, o: Any) -> Any:
    if isinstance(o, FireflyPool):
      return {
          'capacity': o.capacity,
          # pylint: disable=protected-access
          '_last_id': o._last_id,
          '_max_fly_id': o._max_fly_id,
          '_pool': o._pool,
      }
    elif isinstance(o, Firefly):
      return {
          'id_': o.id_,
          'perturbation': o.perturbation,
          'generation': o.generation,
          'trial': o.trial,
      }
    elif isinstance(o, vz.Trial):
      return {
          'parameters':
              o.parameters.as_dict(),
          'objective':
              o.final_measurement.metrics[eagle_strategy_utils.OBJECTIVE_NAME
                                         ].value,
      }
    else:
      return json.JSONEncoder.default(self, o)


@attr.define
class FireflyPoolDecoder:
  """Eagle strategy pool decoder.

  Fully restores the state of the FireflyPool.

  Attributes:
    utils: EagleStrategyUtils initialized with the appropriate random generator.
  """

  _utils: EagleStrategyUtils

  def decode(self, obj: Any) -> FireflyPool:
    """Decodes a string object to partial FireflyPool."""
    obj_dict = json.loads(obj)
    restored_pool = {}

    # Check that all keys appear in restored dictionary.
    missing_keys = set(['_pool', 'capacity', '_last_id', '_max_fly_id']) - set(
        obj_dict.keys())
    if missing_keys:
      raise serializable.HarmlessDecodeError(
          "Couldn't load FireflyPool from metadata. The following keys are "
          'missing: %s' % str(missing_keys))

    # Restore FireFly objects in the pool.
    for id_, fly in obj_dict['_pool'].items():
      trial = vz.Trial(parameters=fly['trial']['parameters'])
      trial.complete(
          measurement=vz.Measurement(
              metrics={'objective': fly['trial'][OBJECTIVE_NAME]}))
      restored_pool[int(id_)] = Firefly(
          id_=fly['id_'],
          perturbation=fly['perturbation'],
          generation=fly['generation'],
          trial=trial,
      )

    restored_capacity = int(obj_dict['capacity'])
    restored_firefly_pool = FireflyPool(
        capacity=restored_capacity, utils=self._utils)
    # pylint: disable=protected-access
    restored_firefly_pool._pool = restored_pool
    restored_firefly_pool._last_id = int(obj_dict['_last_id'])
    restored_firefly_pool._max_fly_id = int(obj_dict['_max_fly_id'])
    return restored_firefly_pool


def partially_serialize_firefly_pool(firefly_pool: FireflyPool) -> str:
  """Serialize parts of the FireflyPool."""
  return json.dumps(firefly_pool, cls=PartialFireflyPoolEncoder)


def restore_firefly_pool(utils: EagleStrategyUtils, obj: str) -> FireflyPool:
  """Fully restore the FireflyPool."""
  return FireflyPoolDecoder(utils).decode(obj)


def serialize_rng(rng: np.random.Generator) -> str:
  """Serialize Numpy Random Genertor."""
  return json.dumps(rng.bit_generator.state)


def restore_rng(obj: str) -> np.random.Generator:
  """Restore Numpy Random Genertor."""
  rng = np.random.default_rng()
  rng.bit_generator.state = json.loads(obj)
  return rng



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for xla_pareto."""

import numpy as np
from vizier._src.jax import xla_pareto
from vizier._src.pyvizier.multimetric import hypervolume

from absl.testing import absltest


class XlaParetoTest(absltest.TestCase):

  def test_get_pareto(self):
    ys = np.random.random([20000, 2])
    is_frontier = xla_pareto.is_frontier(ys, num_shards=5)
    frontier = xla_pareto.get_frontier(ys, num_shards=5)

    # Two implementations should be consistent.
    self.assertEqual(frontier.shape[0], np.sum(is_frontier))

    # Sort by one of the dimensions and the other dimension must be sorted
    # in the reverse order.
    sid = frontier[:, 0].argsort()

    np.testing.assert_array_equal(frontier[sid, 1].argsort(),
                                  np.arange(len(frontier)-1, -1, -1))

  def test_get_pareto_rank(self):
    ys = np.array([[6, 6], [4, 8], [2, 6], [3, 6], [1, 8], [7, 0], [4, 4],
                   [9, 6], [5, 10], [8, 10]])
    np.testing.assert_array_equal(
        xla_pareto.pareto_rank(ys), np.array([2, 2, 6, 5, 3, 2, 5, 0, 1, 0]))

  def test_cum_hypervolume_origin(self):
    dim = 3
    vectors = abs(
        np.array(np.random.uniform(size=(100, dim)), dtype=np.float32))
    points = abs(
        np.array(np.random.uniform(size=(1000, dim)), dtype=np.float32))

    np.testing.assert_array_almost_equal(
        hypervolume._cum_hypervolume_origin(points, vectors),
        xla_pareto.jax_cum_hypervolume_origin(points, vectors),
        decimal=2)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""A meta-Designer to adaptively ensemble multiple Designers."""

from typing import Callable, Optional, Sequence

from absl import logging
import attrs
import numpy as np
from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier._src.algorithms.ensemble import ensemble_design
from vizier.benchmarks import analyzers

ENSEMBLE_NS = 'ens'
EXPERT_KEY = 'expert'


@attrs.define
class ObjectiveRewardGenerator:
  """Stateful rewards generator using the objective curves of Trials."""

  problem: vz.ProblemStatement = attrs.field()
  all_trials: list[vz.Trial] = attrs.field(factory=list)
  reward_regularization: float = attrs.field(
      default=0.0,
      validator=[attrs.validators.instance_of(float), attrs.validators.ge(0)],
  )
  min_reward: float = attrs.field(
      default=0.0,
      validator=[attrs.validators.instance_of(float), attrs.validators.ge(0)],
  )
  # Arguments passed to the hypervolume converter.
  reference_value: Optional[np.ndarray] = attrs.field(
      default=None,
      kw_only=True,
      validator=[
          attrs.validators.optional(attrs.validators.instance_of(np.ndarray))
      ],
  )
  num_vectors: int = attrs.field(
      default=100, kw_only=True, validator=[attrs.validators.ge(0)]
  )

  def __call__(self, trials: list[vz.Trial]) -> list[float]:
    """Generate rewards from trials."""
    # Using MultiMetricConverter for safety understanding.
    if self.problem.is_single_objective:

      def curve_generator() -> analyzers.StatefulCurveConverter:
        return analyzers.MultiMetricCurveConverter.from_metrics_config(
            self.problem.metric_information, flip_signs_for_min=True
        )

    else:

      def curve_generator() -> analyzers.StatefulCurveConverter:
        return analyzers.MultiMetricCurveConverter.from_metrics_config(
            self.problem.metric_information,
            reference_value=self.reference_value,
            num_vectors=self.num_vectors,
            infer_origin_factor=0.1,
        )

    stateful_curve_generator = analyzers.RestartingCurveConverter(
        curve_generator, restart_min_trials=10, restart_rate=1.5
    )

    original_num_trials = len(self.all_trials) - len(trials)
    if original_num_trials <= 0:
      # First round so there is no reward signal.
      rewards = [self.min_reward] * len(trials)
    else:
      # Objective curve is a 1 x (1+len(trials)) array.
      objective_curve = stateful_curve_generator.convert(
          self.all_trials[-1:] + trials
      )
      rewards = []
      for idx in range(len(trials)):
        obj_reward = objective_curve.ys[:, idx + 1] - objective_curve.ys[:, idx]
        regularized_reward = (
            obj_reward
            + self.reward_regularization * objective_curve.ys[:, idx + 1]
        )
        if np.isfinite(regularized_reward):
          rewards.append(
              max(self.min_reward, float(np.squeeze(regularized_reward)))
          )
        else:
          rewards.append(self.min_reward)

    self.all_trials.extend(trials)
    return rewards


class EnsembleDesigner(vza.Designer):
  """Ensembling of Designers."""

  def __init__(
      self,
      designers: dict[str, vza.Designer],
      ensemble_design_factory: Callable[
          [list[int]], ensemble_design.EnsembleDesign
      ] = ensemble_design.RandomEnsembleDesign,
      reward_generator: Optional[ObjectiveRewardGenerator] = None,
      *,
      use_diverse_suggest: bool = False,
      use_separate_update: bool = False,
  ):
    """Creates a wrapper that uses an ensemble of pydesigners.

    Args:
      designers: Dictionary of PyDesigners to be ensembled.
      ensemble_design_factory: EnsembleDesign factory that takes in list of
        indices.
      reward_generator: Generates rewards for use in ensembling strategy.
      use_diverse_suggest: Whether to use a diverse set of designers for batched
        Suggests by repeatedly calling Suggest(1) in round robin fashion.
      use_separate_update: If false, update all Designers with Trials that were
        Suggested by other Designers.

    Raises:
      ValueError when designers is empty.
    """
    if not designers:
      raise ValueError('Empty list of Designers for ensembling.')
    self._designers = designers
    self._ensemble_design_factory = ensemble_design_factory
    # Each index corresponds to the designer order in the dictionary.
    self._strategy = ensemble_design_factory(list(range(len(self._designers))))
    self._use_diverse_suggest = use_diverse_suggest
    self._reward_generator = reward_generator
    self._use_separate_update = use_separate_update

  def suggest(self, num_suggestions: int) -> Sequence[vz.TrialSuggestion]:
    """Randomly chooses a designer and Suggests from the chosen designer.

    Args:
      num_suggestions: Number of suggested trials desired.

    Returns:
      Batch of suggested trials.
    """

    if num_suggestions == 1 or not self._use_diverse_suggest:
      probs = self._strategy.ensemble_probs
      logging.info(
          'Choosing ensemble with probabilities %s over %s',
          probs,
          self._designers.keys(),
      )
      designer_key = np.random.choice(list(self._designers.keys()), p=probs)
      trials = self._designers[designer_key].suggest(num_suggestions)
      for t in trials:
        t.metadata.ns(ENSEMBLE_NS)[EXPERT_KEY] = f'{designer_key}'
      return trials
    else:
      # Apply diverse Suggestions.
      diverse_suggestions = []
      for _ in range(num_suggestions):
        diverse_suggestions.extend(self.suggest(1))
      return diverse_suggestions

  def update(
      self, completed: vza.CompletedTrials, all_active: vza.ActiveTrials
  ) -> None:
    # With no completed Trials, simply update with reward calcuations.
    if not completed.trials:
      for designer in self._designers.values():
        designer.update(completed, all_active)
      return

    if self._reward_generator is None:
      rewards = []
      for t in completed.trials:
        if t.final_measurement is not None:
          metrics = t.final_measurement.metrics
          first_key = list(metrics.keys())[0]
          rewards.append(metrics.get_value(first_key, default=0.0))
    else:
      rewards = self._reward_generator(list(completed.trials))

    # Update the EnsembleDesign strategy.
    for trial, reward in zip(completed.trials, rewards):
      designer_key = trial.metadata.ns(ENSEMBLE_NS).get(EXPERT_KEY)
      if designer_key is None:
        logging.warning(
            'Trial does not contain required metadata with key %s: %s',
            EXPERT_KEY,
            trial,
        )
        continue

      observation = (list(self._designers.keys()).index(designer_key), reward)
      logging.info('Updating with expert, rewards: %s', observation)
      self._strategy.update(observation)

    # Update the underlying designers.
    if self._use_separate_update:
      for designer_key, designer in self._designers.items():
        filtered_completed = [
            t
            for t in completed.trials
            if t.metadata.ns(ENSEMBLE_NS).get(EXPERT_KEY) == designer_key
        ]
        filtered_active = [
            t
            for t in all_active.trials
            if t.metadata.ns(ENSEMBLE_NS).get(EXPERT_KEY) == designer_key
        ]
        designer.update(
            vza.CompletedTrials(filtered_completed),
            vza.ActiveTrials(filtered_active),
        )
    else:
      for designer in self._designers.values():
        designer.update(completed, all_active)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tools for visualization with PlotElements."""


import json
from typing import Optional, Sequence, Tuple
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from vizier import pyvizier as vz
from vizier._src.benchmarks.analyzers import state_analyzer
from vizier.utils import json_utils


def plot_median_convergence(
    ax: mpl.axes.Axes,
    curves: 'np.ndarray',
    *,
    percentiles: Sequence[Tuple[int, int]] = ((40, 60),),
    alphas: Sequence[float] = (0.2,),
    xs: Optional['np.ndarray'] = None,
    **kwargs,
):
  """Aggregates multiple convergence curves into a plot with confidence bounds.

  Example usage:
    ```python
    fig, ax = plt.subplots(1, 1, figsize=(12,8))
    plot_median_convergence(ax,
                            [[1,1,2,3,4], [1,1,1,2,nan]],
                            percentiles=((40, 60), (30, 70)),
                            alphas=(0.4, 0.2),
                            xs=np.arange(1,6),
                            color='r')
    ```

  Args:
    ax: matplotlib axis to plot on.
    curves: Expected to have shape (Number of studies, points), where rows are
      convergence curves from repeated studies from the same algorithm and
      settings. May contain NaNs, which will be excluded from plotting.
    percentiles: Each pair defines (lower_percentile, upper_percentile).
    alphas: Must have the same length as percentiles. Defines the color strength
      for the confidence bounds. Make it decrease in the distance between lower
      and upper percentiles. (See example above).
    xs: x values for the plot. If not provided, uses np.arange(curves.shape[1]).
      Must have the shape (curves.shape[1], 0)
    **kwargs: Forwared to ax.plot().
  """
  if xs is None:
    xs = np.arange(curves.shape[1])

  line = ax.plot(xs, np.nanmedian(curves, axis=0), **kwargs)
  for (lower, upper), alpha in zip(percentiles, alphas):
    ax.fill_between(
        xs,
        np.nanpercentile(curves, lower, axis=0),
        np.nanpercentile(curves, upper, axis=0),
        alpha=alpha,
        color=line[0].get_color(),
    )


def plot_mean_convergence(
    ax: mpl.axes.Axes,
    curves: 'np.ndarray',
    *,
    alpha: float = 0.2,
    xs: Optional['np.ndarray'] = None,
    **kwargs,
):
  """Aggregates multiple convergence curves into a plot with standard error bounds.

  Example usage:
    ```python
    fig, ax = plt.subplots(1, 1, figsize=(12,8))
    plot_mean_convergence(ax,
                          [[1,1,2,3,4], [1,1,1,2,nan]],
                          alpha=0.3,
                          xs=np.arange(1,6),
                          color='r')
    ```

  Args:
    ax: matplotlib axis to plot on.
    curves: Expected to have shape (Number of studies, points), where rows are
      convergence curves from repeated studies from the same algorithm and
      settings. May contain NaNs, which will be excluded from plotting.
    alpha: Defines the color strength for the standard error bounds.
    xs: x values for the plot. If not provided, uses np.arange(curves.shape[1]).
      Must have the shape (curves.shape[1], 0)
    **kwargs: Forwared to ax.plot().
  """
  if xs is None:
    xs = np.arange(curves.shape[1])
  curves_mean = np.nanmean(curves, axis=0)
  curves_std_error = np.nanstd(curves, axis=0) / np.sqrt(curves.shape[0])

  line = ax.plot(xs, curves_mean, **kwargs)
  ax.fill_between(
      xs,
      curves_mean + 1.5 * curves_std_error,
      curves_mean - 1.5 * curves_std_error,
      alpha=alpha,
      color=line[0].get_color(),
  )


def plot_from_records(
    records: Sequence[state_analyzer.BenchmarkRecord],
    metrics: Optional[Sequence[str]] = None,
    *,
    fig_title: str = 'All Plot Elements',
    title_maxlen: int = 50,
    col_figsize: float = 6.0,
    row_figsize: float = 6.0,
    **kwargs,
):
  """Generates a grid of algorithm comparison plots.

  Generates one plot for each Experimenter x Metrics in records. Note that
  each row = Experimenter and each column = Metrics.

  Args:
    records: All BenchmarkRecords used for plotting.
    metrics: Keys in the plot_elements dict in BenchmarkRecord used for plot. If
      not supplied, all keys are plotted.
    fig_title: Title of the entire grid plot.
    title_maxlen: Maximum length of title of each Experimenter.
    col_figsize: Size of the column of each subfigure.
    row_figsize: Size of the row of each subfigure.
    **kwargs: Additional keyword args forwarded to pyplot.

  Raises:
    ValueError: When plot type is not supported.
  """

  def _metadata_to_str(metadata: vz.Metadata) -> str:
    visual_dict = {}
    for _, key, value in metadata.all_items():
      try:
        loaded = json.loads(value, cls=json_utils.NumpyDecoder)
        assert isinstance(loaded, dict)
        visual_dict = visual_dict | {k: v for k, v in loaded.items() if v}
      except Exception as e:  # pylint: disable=broad-except
        del e
        visual_dict[key] = value
    return str(visual_dict)

  records_list = [
      (rec.algorithm, _metadata_to_str(rec.experimenter_metadata), rec)
      for rec in records
  ]
  df = pd.DataFrame(
      records_list, columns=['algorithm', 'experimenter', 'record']
  )

  algorithms = df.algorithm.unique()
  colors = {
      algorithm: plt.get_cmap('tab10')(i)
      for i, algorithm in enumerate(algorithms)
  }
  total_rows = len(df.groupby('experimenter'))
  if metrics is None:
    metrics = set()
    for record in df.record:
      metrics = metrics.union(set(record.plot_elements.keys()))
    print(f'All inferred metrics {metrics}')

  fig, axes = plt.subplots(
      total_rows,
      len(metrics),
      figsize=(col_figsize * len(metrics), row_figsize * total_rows),
      squeeze=False,
  )
  fig.suptitle(fig_title, fontsize=16)

  fig_idx = 0
  for experimenter_key, group_by_experimenter in df.groupby('experimenter'):
    for metric_idx, metric in enumerate(metrics):
      ax = axes[fig_idx, metric_idx]
      subplot_title = str(experimenter_key)[:title_maxlen]
      ax.set_title(subplot_title)
      ax.set_ylabel(metric)
      for algorithm_name, group in group_by_experimenter.groupby('algorithm'):
        if not group.size:
          continue
        if len(group) != 1:
          print(
              f'Found more records than expected in {algorithm_name} for'
              f' {group}'
          )
        elems = group.record.iloc[0].plot_elements
        if metric not in elems:
          print(f'metric {metric} not found in {group.record.iloc[0]}')
          continue

        elem_for_metric = elems[metric]
        plot_type = elem_for_metric.plot_type
        if plot_type == 'error-bar':
          plot_median_convergence(
              ax,
              elem_for_metric.curve.ys,
              xs=elem_for_metric.curve.xs,
              label=f'{algorithm_name}',
              color=colors[algorithm_name],
              percentiles=(elem_for_metric.percentile_error_bar,),
              **kwargs,
          )
          ax.set_xlabel('# of Trials')
        elif plot_type == 'scatter':
          plot = elem_for_metric.plot_array
          ax.scatter(
              plot[:, 0],
              plot[:, 1],
              label=f'{algorithm_name}',
              color=colors[algorithm_name],
              **kwargs,
          )
        elif plot_type == 'histogram':
          plot = elem_for_metric.plot_array
          linewidth = (
              len(algorithms)
              + 1
              - float(list(algorithms).index(algorithm_name))
          )
          ax.hist(
              plot,
              histtype='step',
              density=True,
              fill=False,
              linewidth=linewidth,
              label=f'{algorithm_name}',
              color=colors[algorithm_name],
              **kwargs,
          )
        else:
          raise ValueError(f'{plot_type} plot not yet supported!')
        ax.set_yscale(elem_for_metric.yscale)
        ax.yaxis.set_major_locator(mpl.ticker.LinearLocator(20))
        ax.yaxis.set_minor_locator(mpl.ticker.LinearLocator(100))
        ax.yaxis.set_major_formatter(mpl.ticker.ScalarFormatter())
      ax.legend()
    fig_idx += 1



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Collection of well-tuned GP models."""

# TODO: Add Ax/BoTorch GP.

import functools
from typing import Any, Generator, Optional, Union

from flax import struct
import jax
from jax import numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp
from vizier._src.jax import stochastic_process_model as sp
from vizier._src.jax import types
from vizier._src.jax.models import continuous_only_kernel
from vizier._src.jax.models import mask_features

tfb = tfp.bijectors
tfd = tfp.distributions
tfpk = tfp.math.psd_kernels
tfpke = tfp.experimental.psd_kernels


def _log_uniform_init(
    low: Union[float, np.floating],
    high: Union[float, np.floating],
    shape: tuple[int, ...] = tuple(),
) -> sp.InitFn:
  r"""Take log-uniform sample in the constraint and map it back to \R.

  Args:
    low: Parameter lower bound.
    high: Parameter upper bound.
    shape: Returned array has this shape. Each entry in the returned array is an
      i.i.d sample.

  Returns:
    Randomly sampled array.
  """

  def sample(key: Any) -> jnp.ndarray:
    unif = jax.random.uniform(key, shape, dtype=jnp.float64)
    return jnp.exp(unif * jnp.log(high / low) + jnp.log(low))

  return sample


@struct.dataclass
class VizierGaussianProcess(sp.ModelCoroutine[tfd.GaussianProcess]):
  """Vizier's tuned GP with categorical parameters.

  See __call__ method documentation.

  Attributes:
    _boundary_epsilon: We expand the constraints by this number so that the
      values exactly at the boundary can be mapped to unconstrained space. i.e.
      we are trying to avoid SoftClip(low=1e-2, high=1.).inverse(1e-2) giving
      NaN.
  """

  _dim: types.ContinuousAndCategorical[int] = struct.field(pytree_node=False)
  _use_retrying_cholesky: bool = struct.field(
      pytree_node=False, default=True, kw_only=True
  )
  _boundary_epsilon: float = struct.field(default=1e-12, kw_only=True)
  _linear_coef: float = struct.field(default=1e-12, kw_only=True)

  @classmethod
  def build_model(
      cls,
      features: types.ModelInput,
      *,
      use_retrying_cholesky: bool = True,
      linear_coef: float = 1e-12,
  ) -> sp.StochasticProcessModel:
    """Returns the model and loss function."""
    gp_coroutine = VizierGaussianProcess(
        _dim=types.ContinuousAndCategorical[int](
            features.continuous.padded_array.shape[-1],
            features.categorical.padded_array.shape[-1],
        ),
        _use_retrying_cholesky=use_retrying_cholesky,
        _linear_coef=linear_coef,
    )
    return sp.StochasticProcessModel(gp_coroutine)

  def __call__(
      self,
      inputs: Optional[types.ModelInput] = None,
  ) -> Generator[sp.ModelParameter, jax.Array, tfd.GaussianProcess]:
    """Creates a generator.

    Args:
      inputs: ContinuousAndCategoricalArray with array shapes of: (num_examples,
        continuous_feature_dim), (num_examples, categorical_feature_dim).

    Yields:
      GaussianProcess whose event shape is `num_examples`.
    """
    eps = self._boundary_epsilon
    observation_noise_bounds = (np.float64(1e-10 - eps), 1.0 + eps)
    amplitude_bounds = (np.float64(1e-3 - eps), 10.0 + eps)
    continuous_ones = np.ones((self._dim.continuous), dtype=np.float64)
    continuous_length_scale_bounds = (
        continuous_ones * (1e-2 - eps),
        continuous_ones * 1e2 + eps,
    )
    categorical_ones = np.ones((self._dim.categorical), dtype=np.float64)
    categorical_length_scale_bounds = (
        categorical_ones * (1e-2 - eps),
        categorical_ones * 1e2 + eps,
    )

    signal_variance = yield sp.ModelParameter(
        init_fn=_log_uniform_init(*amplitude_bounds),
        constraint=sp.Constraint(
            amplitude_bounds,
            tfb.SoftClip(*amplitude_bounds, hinge_softness=1e-2),
        ),
        regularizer=lambda x: 0.01 * jnp.log(x / 0.039) ** 2,
        name='signal_variance',
    )
    kernel = tfpk.MaternFiveHalves(amplitude=jnp.sqrt(signal_variance))

    continuous_length_scale_squared = yield sp.ModelParameter(
        init_fn=_log_uniform_init(
            *continuous_length_scale_bounds, shape=(self._dim.continuous,)
        ),
        constraint=sp.Constraint(
            continuous_length_scale_bounds,
            tfb.SoftClip(*continuous_length_scale_bounds, hinge_softness=1e-2),
        ),
        regularizer=lambda x: jnp.sum(0.01 * jnp.log(x / 0.5) ** 2),
        name='continuous_length_scale_squared',
    )
    categorical_length_scale_squared = yield sp.ModelParameter(
        init_fn=_log_uniform_init(
            *categorical_length_scale_bounds,
            shape=(self._dim.categorical,),
        ),
        constraint=sp.Constraint(
            categorical_length_scale_bounds,
            tfb.SoftClip(*categorical_length_scale_bounds, hinge_softness=1e-2),
        ),
        regularizer=lambda x: jnp.sum(0.01 * jnp.log(x / 0.5) ** 2),
        name='categorical_length_scale_squared',
    )
    kernel = tfpke.FeatureScaledWithCategorical(
        kernel,
        scale_diag=tfpke.ContinuousAndCategoricalValues(
            jnp.sqrt(continuous_length_scale_squared),
            jnp.sqrt(categorical_length_scale_squared),
        ),
    )
    if inputs is not None:
      # Ensure features are zero for this kernel. This will also ensure the
      # length scales are not trainable, since there will be no signal from
      # these dimensions.
      kernel = mask_features.MaskFeatures(
          kernel,
          dimension_is_missing=tfpke.ContinuousAndCategoricalValues(
              continuous=inputs.continuous.is_missing[1],
              categorical=inputs.categorical.is_missing[1],
          ),
      )
      inputs = tfpke.ContinuousAndCategoricalValues(
          continuous=inputs.continuous.padded_array,
          categorical=inputs.categorical.padded_array,
      )

    observation_noise_variance = yield sp.ModelParameter(
        init_fn=_log_uniform_init(*observation_noise_bounds),
        constraint=sp.Constraint(
            observation_noise_bounds,
            tfb.SoftClip(*observation_noise_bounds, hinge_softness=1e-2),
        ),
        regularizer=lambda x: 0.01 * jnp.log(x / 0.0039) ** 2,
        name='observation_noise_variance',
    )
    cholesky_fn = None
    # When cholesky fails, increase jitters and retry.
    if self._use_retrying_cholesky:
      retrying_cholesky = functools.partial(
          tfp.experimental.distributions.marginal_fns.retrying_cholesky,
          jitter=np.float64(1e-4),
          max_iters=5,
      )
      cholesky_fn = lambda matrix: retrying_cholesky(matrix)[0]

    return tfd.GaussianProcess(
        kernel,
        index_points=inputs,
        observation_noise_variance=observation_noise_variance,
        cholesky_fn=cholesky_fn,
    )


@struct.dataclass
class ConstantMeanFn:
  """Implements a constant GP mean function."""

  constant: jax.Array

  def __call__(self, x: types.ModelInput) -> jax.Array:
    del x
    return self.constant


# TODO: Consider renaming this to reflect that it uses a
# Matern + Linear kernel, not just a Linear kernel (or merging this class with
# VizierGaussianProcess).
@struct.dataclass
class VizierLinearGaussianProcess(sp.ModelCoroutine[tfd.GaussianProcess]):
  """Vizier's tuned GP with linear kernel.

  See __call__ method documentation.

  Attributes:
    _boundary_epsilon: We expand the constraints by this number so that the
      values exactly at the boundary can be mapped to unconstrained space. i.e.
      we are trying to avoid SoftClip(low=1e-2, high=1.).inverse(1e-2) giving
      NaN.
  """

  _dim: types.ContinuousAndCategorical[int] = struct.field(pytree_node=False)
  _use_retrying_cholesky: bool = struct.field(
      pytree_node=False, default=True, kw_only=True
  )
  _boundary_epsilon: float = struct.field(default=1e-12, kw_only=True)
  _linear_coef: float = struct.field(default=1e-12, kw_only=True)

  # TODO: Redesign or remove this method given the updated
  # StochasticProcessModelCoroutine class.
  @classmethod
  def build_model(
      cls,
      features: types.ModelInput,
      *,
      use_retrying_cholesky: bool = True,
      linear_coef: float = 1e-12,
  ) -> sp.StochasticProcessModel:
    """Returns the model and loss function."""
    gp_coroutine = VizierLinearGaussianProcess(
        _dim=types.ContinuousAndCategorical[int](
            features.continuous.padded_array.shape[-1],
            features.categorical.padded_array.shape[-1],
        ),
        _use_retrying_cholesky=use_retrying_cholesky,
        _linear_coef=linear_coef,
    )
    return sp.StochasticProcessModel(gp_coroutine)

  def __call__(
      self,
      inputs: Optional[types.ModelInput] = None,
  ) -> Generator[sp.ModelParameter, jax.Array, tfd.GaussianProcess]:
    """Creates a generator.

    Args:
      inputs: ContinuousAndCategoricalArray with array shapes of: (num_examples,
        continuous_feature_dim), (num_examples, categorical_feature_dim).

    Yields:
      GaussianProcess whose event shape is `num_examples`.
    """
    eps = self._boundary_epsilon
    observation_noise_bounds = (np.float64(1e-10 - eps), 1.0 + eps)
    amplitude_bounds = (np.float64(1e-3 - eps), 10.0 + eps)
    continuous_ones = np.ones((self._dim.continuous), dtype=np.float64)
    continuous_length_scale_bounds = (
        continuous_ones * (1e-2 - eps),
        continuous_ones * 1e2 + eps,
    )
    categorical_ones = np.ones((self._dim.categorical), dtype=np.float64)
    categorical_length_scale_bounds = (
        categorical_ones * (1e-2 - eps),
        categorical_ones * 1e2 + eps,
    )

    signal_variance = yield sp.ModelParameter(
        init_fn=_log_uniform_init(*amplitude_bounds),
        constraint=sp.Constraint(
            amplitude_bounds,
            tfb.SoftClip(*amplitude_bounds, hinge_softness=1e-2),
        ),
        regularizer=lambda x: 0.01 * jnp.log(x / 0.039) ** 2,
        name='signal_variance',
    )
    kernel = tfpk.MaternFiveHalves(amplitude=jnp.sqrt(signal_variance))

    continuous_length_scale_squared = yield sp.ModelParameter(
        init_fn=_log_uniform_init(
            *continuous_length_scale_bounds, shape=(self._dim.continuous,)
        ),
        constraint=sp.Constraint(
            continuous_length_scale_bounds,
            tfb.SoftClip(*continuous_length_scale_bounds, hinge_softness=1e-2),
        ),
        regularizer=lambda x: jnp.sum(0.01 * jnp.log(x / 0.5) ** 2),
        name='continuous_length_scale_squared',
    )
    categorical_length_scale_squared = yield sp.ModelParameter(
        init_fn=_log_uniform_init(
            *categorical_length_scale_bounds,
            shape=(self._dim.categorical,),
        ),
        constraint=sp.Constraint(
            categorical_length_scale_bounds,
            tfb.SoftClip(*categorical_length_scale_bounds, hinge_softness=1e-2),
        ),
        regularizer=lambda x: jnp.sum(0.01 * jnp.log(x / 0.5) ** 2),
        name='categorical_length_scale_squared',
    )
    kernel = tfpke.FeatureScaledWithCategorical(
        kernel,
        scale_diag=tfpke.ContinuousAndCategoricalValues(
            jnp.sqrt(continuous_length_scale_squared),
            jnp.sqrt(categorical_length_scale_squared),
        ),
    )

    # We do not need to tune bias here because we are tuning mean_fn.
    slopes = yield sp.ModelParameter(
        init_fn=_log_uniform_init(*amplitude_bounds),
        constraint=sp.Constraint(
            amplitude_bounds,
            tfb.SoftClip(*amplitude_bounds, hinge_softness=1e-2),
        ),
        regularizer=lambda x: 0.01 * jnp.log(x / 0.039) ** 2,
        name='slope_amplitude',
    )
    shift = yield sp.ModelParameter(
        init_fn=jax.random.normal,
        regularizer=lambda x: 0.5 * x**2,
        name='shift',
    )
    kernel = kernel + continuous_only_kernel.ContinuousOnly(
        tfpk.FeatureScaled(
            tfpk.Linear(
                slope_amplitude=self._linear_coef * slopes,
                shift=self._linear_coef * shift,
            ),
            scale_diag=jnp.sqrt(continuous_length_scale_squared),
        )
    )

    if inputs is not None:
      # Ensure features are zero for this kernel. This will also ensure the
      # length scales are not trainable, since there will be no signal from
      # these dimensions.
      kernel = mask_features.MaskFeatures(
          kernel,
          dimension_is_missing=tfpke.ContinuousAndCategoricalValues(
              continuous=inputs.continuous.is_missing[1],
              categorical=inputs.categorical.is_missing[1],
          ),
      )
      inputs = tfpke.ContinuousAndCategoricalValues(
          continuous=inputs.continuous.padded_array,
          categorical=inputs.categorical.padded_array,
      )

    observation_noise_variance = yield sp.ModelParameter(
        init_fn=_log_uniform_init(*observation_noise_bounds),
        constraint=sp.Constraint(
            observation_noise_bounds,
            tfb.SoftClip(*observation_noise_bounds, hinge_softness=1e-2),
        ),
        regularizer=lambda x: 0.01 * jnp.log(x / 0.0039) ** 2,
        name='observation_noise_variance',
    )
    cholesky_fn = None
    # When cholesky fails, increase jitters and retry.
    if self._use_retrying_cholesky:
      retrying_cholesky = functools.partial(
          tfp.experimental.distributions.marginal_fns.retrying_cholesky,
          jitter=np.float64(1e-4),
          max_iters=5,
      )
      cholesky_fn = lambda matrix: retrying_cholesky(matrix)[0]

    # The mean function output must broadcast to
    # `[batch_shape, num_observations]`, where `batch_shape` is the
    # (possibly empty) batch shape of the other hyperparameters. Initializing
    # `mean_fn_constant` with an array of shape `[1]` gives the mean function
    # output a shape of `[batch_shape, 1]`, ensuring that batch dimensions line
    # up properly.
    mean_fn_constant = yield sp.ModelParameter(
        init_fn=lambda k: jax.random.normal(key=k, shape=[1]),
        regularizer=lambda x: 0.5 * jnp.squeeze(x, axis=-1) ** 2,
        name='mean_fn',
    )

    mean_fn = ConstantMeanFn(mean_fn_constant * self._linear_coef)

    return tfd.GaussianProcess(
        kernel,
        index_points=inputs,
        observation_noise_variance=observation_noise_variance,
        cholesky_fn=cholesky_fn,
        mean_fn=mean_fn,
    )



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Test suite for SparseExperimenter."""

import attr
from vizier import pyvizier as vz
from vizier._src.benchmarks.experimenters import experimenter_factory
from vizier._src.benchmarks.experimenters import sparse_experimenter

from absl.testing import absltest
from absl.testing import parameterized


class SparseExperimenterTest(parameterized.TestCase):

  @parameterized.parameters(
      vz.ParameterConfig.factory('_', bounds=(0.0, 5.0)),
      vz.ParameterConfig.factory('_', bounds=(0, 5)),
      vz.ParameterConfig.factory('_', feasible_values=[1, 2, 3]),
      vz.ParameterConfig.factory('_', feasible_values=['a', 'b', 'c']),
  )
  def test_sparse_problem_statement_and_evaluate(self, sparse_param_config):
    sphere_experimenter = experimenter_factory.BBOBExperimenterFactory(
        name='Sphere', dim=2
    )()
    search_space = vz.SearchSpace()
    for idx in range(8):
      search_space.add(attr.evolve(sparse_param_config, name=str(idx)))

    experimenter = sparse_experimenter.SparseExperimenter(
        experiment=sphere_experimenter,
        search_space=search_space,
        prefix='__SPARSE',
    )
    # Test that sparse parameters were added.
    self.assertLen(experimenter.problem_statement().search_space.parameters, 10)
    trial = vz.Trial()
    for pc in experimenter.problem_statement().search_space.parameters:
      trial.parameters[pc.name] = 2.0
    # Test that the evaluation uses only the non-sparse parameters.
    experimenter.evaluate([trial])
    self.assertEqual(trial.final_measurement.metrics['bbob_eval'].value, 8.0)
    # Test that the evaluated trial parameters remained the same.
    self.assertLen(trial.parameters, 10)
    for trial_param_name, exptr_param_config in zip(
        trial.parameters,
        sphere_experimenter.problem_statement().search_space.parameters,
    ):
      self.assertEqual(trial_param_name, exptr_param_config.name)

  @parameterized.parameters(
      {
          'float_count': 1,
          'int_count': 2,
          'discrete_count': 4,
          'cat_count': 3,
      },
      {'float_count': 10, 'int_count': 0, 'discrete_count': 0, 'cat_count': 0},
      {'float_count': 0, 'int_count': 0, 'discrete_count': 0, 'cat_count': 0},
      {'float_count': 0, 'int_count': 0, 'discrete_count': 3, 'cat_count': 5},
  )
  def test_sparse_experimenter_create(
      self, float_count, int_count, discrete_count, cat_count
  ):
    sphere_experimenter = experimenter_factory.BBOBExperimenterFactory(
        name='Sphere', dim=2
    )()
    experimenter = sparse_experimenter.SparseExperimenter.create(
        experiment=sphere_experimenter,
        float_count=float_count,
        int_count=int_count,
        discrete_count=discrete_count,
        categorical_count=cat_count,
    )

    # Test that sparse parameters were added corretly to the search space.
    def _count_param_by_type(problem):
      """Count the number of parameters by parameter type."""
      counts = {type: 0 for type in vz.ParameterType}
      for param in problem.search_space.parameters:
        counts[param.type] += 1
      return counts

    counts = _count_param_by_type(experimenter.problem_statement())
    self.assertEqual(counts[vz.ParameterType.DOUBLE], float_count + 2)
    self.assertEqual(counts[vz.ParameterType.INTEGER], int_count)
    self.assertEqual(counts[vz.ParameterType.CATEGORICAL], cat_count)
    self.assertEqual(counts[vz.ParameterType.DISCRETE], discrete_count)

    trial = vz.Trial()
    for pc in experimenter.problem_statement().search_space.parameters:
      trial.parameters[pc.name] = 2.0
    # Test that the evaluation uses only the non-sparse parameters.
    experimenter.evaluate([trial])
    self.assertEqual(trial.final_measurement.metrics['bbob_eval'].value, 8.0)
    # Test that the evaluated trial parameters remained the same.
    total_param_count = 2 + int_count + cat_count + discrete_count + float_count
    self.assertLen(trial.parameters, total_param_count)
    for trial_param_name, exptr_param_config in zip(
        trial.parameters,
        sphere_experimenter.problem_statement().search_space.parameters,
    ):
      self.assertEqual(trial_param_name, exptr_param_config.name)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Coroutine for a trainable Gaussian process with an ARD kernel."""

from typing import Any, Generator, Optional, Type, Union

import jax
from jax import numpy as jnp
from tensorflow_probability.substrates import jax as tfp
from vizier._src.jax import stochastic_process_model as sp_model
from vizier._src.jax import types

Array = Any
tfd = tfp.distributions
tfpk = tfp.math.psd_kernels
tfpke = tfp.experimental.psd_kernels


class GaussianProcessARD(sp_model.ModelCoroutine):
  """Specifies an ARD Gaussian process model over continuous/categorical data.

  While `GaussianProcessARD` operates on a continuous domain,
  `GaussianProcessARDWithCategorical` operates on a domain that contains both
  continuous and categorical features. The GP's kernel, a
  `FeatureScaledWithCategorical` instance, operates on inputs of this type as
  well.

  `GaussianProcessARDWithCategorical` satisfies the `ModelCoroutine`
  Protocol. By default, `GaussianProcessARDWithCategorical` uses a Matern
  Five Halves kernel. Hyperparameter priors are LogNormal(0., 1.).

  #### Examples

  Build a Gaussian Process model with an ARD Exponentiated Quadratic kernel and
  LogNormal(0., 1.) hyperparameter priors for 5 continuous features and 3
  categorical features.

  ```python
  from tensorflow_probability.substrates import jax as tfp

  tfpk = tfp.math.psd_kernels
  tfpke = tfp.experimental.psd_kernels

  gen = GaussianProcessARDWithCategorical(
      dimension=types.ContinuousAndCategorical[int](
          continuous=5, categorical=3),
      kernel_class=tfpk.ExponentiatedQuadratic)
  ```
  """

  def __init__(
      self,
      dimension: types.ContinuousAndCategorical[int],
      kernel_class: Union[
          Type[tfpk.MaternFiveHalves],
          Type[tfpk.MaternThreeHalves],
          Type[tfpk.MaternOneHalf],
          Type[tfpk.Parabolic],
          Type[tfpk.ExponentiatedQuadratic],
      ] = tfpk.MaternFiveHalves,
      *,
      use_tfp_runtime_validation: bool = False,
      dtype: jnp.dtype = jnp.float64,
  ):
    """Initializes a `GaussianProcessARDWithCategorical`.

    Args:
      dimension: The dimensionality of the continuous/categorical features.
      kernel_class: The Gaussian process' kernel type.
      use_tfp_runtime_validation: If True, run additional runtime checks on the
        validity of the parameters and input for TFP objects (e.g. verify that
        amplitude and length scale parameters are positive). Runtime checks may
        be expensive and should be used only during development/debugging.
      dtype: Float dtype.
    """
    self.dimension = dimension
    self._kernel_class = kernel_class
    self._use_tfp_runtime_validation = use_tfp_runtime_validation
    self.dtype = dtype
    if dtype == jnp.float64 and not jax.config.read('jax_enable_x64'):
      raise ValueError(
          "x64 is not enabled for jax. Add jax_config.update('jax_enable_x64',"
          ' True) to your main'
      )

  def __call__(
      self, inputs: Optional[types.ModelInput] = None
  ) -> Generator[sp_model.ModelParameter, Array, tfd.GaussianProcess]:
    # TODO: Remove the following line when the linter bug is fixed.
    # pylint: disable=g-doc-return-or-yield
    """The coroutine that specifies the GP model.

    The inputs (and the index points of the returned Gaussian process) are
    instances of `tfpke.ContinuousAndCategoricalValues`, a `namedtuple` with
    a `"continuous"` field that contains continuous (float) data and a
    `"categorical"` field that contains categorical (int) data. The
    categorical data is encoded as integers (not one-hot encoded).

    Args:
      inputs: index_points to be provided to the GP.

    Yields:
      `ModelParameter`s describing the parameters to be declared in the Flax
        model.

    Returns:
      A `tfd.GaussianProcess` with the given index points.
    """
    # pytype: disable=not-callable  # jnp-type
    amplitude = yield sp_model.ModelParameter.from_prior(
        tfd.LogNormal(self.dtype(0.0), 1.0, name='amplitude'),
        constraint=sp_model.Constraint(bounds=(self.dtype(0.0), None)),
    )
    kernel = self._kernel_class(
        amplitude=amplitude,
        length_scale=self.dtype(1.0),
        validate_args=self._use_tfp_runtime_validation,
    )
    inverse_length_scale_continuous = yield sp_model.ModelParameter.from_prior(
        tfd.Sample(
            tfd.LogNormal(self.dtype(0.0), 1.0),
            sample_shape=(self.dimension.continuous,),
            name='inverse_length_scale_continuous',
        ),
        constraint=sp_model.Constraint(bounds=(self.dtype(0.0), None)),
    )
    inverse_length_scale_categorical = yield sp_model.ModelParameter.from_prior(
        tfd.Sample(
            tfd.LogNormal(self.dtype(0.0), 1.0),
            sample_shape=(self.dimension.categorical,),
            name='inverse_length_scale_categorical',
        ),
        constraint=sp_model.Constraint(
            bounds=(jnp.zeros(self.dimension.categorical, dtype=self.dtype),
                    None)
        ),
    )
    kernel = tfpke.FeatureScaledWithCategorical(
        kernel,
        inverse_scale_diag=tfpke.ContinuousAndCategoricalValues(
            inverse_length_scale_continuous, inverse_length_scale_categorical
        ),
        validate_args=self._use_tfp_runtime_validation,
    )
    observation_noise_variance = yield sp_model.ModelParameter.from_prior(
        tfd.LogNormal(self.dtype(0.0), 1.0,
                      name='observation_noise_variance'),
        constraint=sp_model.Constraint(bounds=(self.dtype(0.0), None)),
    )
    # pytype: enable=not-callable
    if inputs is not None:
      inputs = tfpke.ContinuousAndCategoricalValues(
          inputs.continuous.padded_array, inputs.categorical.padded_array
      )
    return tfd.GaussianProcess(
        kernel,
        index_points=inputs,
        observation_noise_variance=observation_noise_variance,
        validate_args=self._use_tfp_runtime_validation,
    )



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for vizier.service.vizier_client."""

# TODO: Rewrite using RPC calls and remove all direct lookups
# into the datastore.
# TODO: Cover delete_trial, delete_study, get_study_config, and
# add_trial, OR turn it into a private module

from typing import List

from absl import logging
from vizier._src.service import constants
from vizier._src.service import resources
from vizier._src.service import study_pb2
from vizier._src.service import vizier_client
from vizier._src.service import vizier_server
from vizier._src.service import vizier_service_pb2_grpc
from vizier.service import pyvizier as vz

from absl.testing import absltest
from absl.testing import parameterized


class VizierClientTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    # Setup Vizier Server and some pre-stored data.
    self.local_server = vizier_server.DefaultVizierServer(
        database_url=constants.SQL_MEMORY_URL
    )
    self.servicer = self.local_server._servicer
    self.owner_id = 'my_username'
    self.study_id = '1231232'
    self.study_resource_name = resources.StudyResource(
        self.owner_id, self.study_id
    ).name

    # Setup connection to server.
    vizier_client.environment_variables.server_endpoint = (
        self.local_server.endpoint
    )
    self.client = vizier_client.VizierClient(
        study_resource_name=self.study_resource_name, client_id='my_client'
    )

    # Store initial data in the vizier service.
    double_value_spec = study_pb2.StudySpec.ParameterSpec.DoubleValueSpec(
        min_value=-1.0, max_value=1.0
    )
    double_parameter_spec = study_pb2.StudySpec.ParameterSpec(
        parameter_id='double', double_value_spec=double_value_spec
    )
    metric_spec = study_pb2.StudySpec.MetricSpec(
        metric_id='example_metric',
        goal=study_pb2.StudySpec.MetricSpec.GoalType.MAXIMIZE,
    )
    self.example_study = study_pb2.Study(
        name=self.study_resource_name,
        study_spec=study_pb2.StudySpec(
            algorithm='RANDOM_SEARCH',
            parameters=[double_parameter_spec],
            metrics=[metric_spec],
        ),
    )
    self.active_trial = study_pb2.Trial(
        name=resources.TrialResource(self.owner_id, self.study_id, 1).name,
        id='1',
        state=study_pb2.Trial.State.ACTIVE,
    )
    self.servicer.datastore.create_study(self.example_study)
    self.servicer.datastore.create_trial(self.active_trial)

  def test_create_or_load_study(self):
    study_config = vz.StudyConfig()
    study_id = 'example_display_name'

    client = vizier_client.create_or_load_study(
        owner_id=self.owner_id,
        client_id='a_client',
        study_id=study_id,
        study_config=study_config,
    )
    study = self.servicer.datastore.load_study(client.study_resource_name)
    self.assertEqual(study.study_spec, study_config.to_proto())
    self.assertIsNotNone(study.name)

    another_client = vizier_client.create_or_load_study(
        owner_id=self.owner_id,
        client_id='another_client',
        study_id=study_id,
        study_config=study_config,
    )
    self.assertEqual(
        client.study_resource_name, another_client.study_resource_name
    )

  def test_list_studies(self):
    study_list_json = self.client.list_studies()
    self.assertLen(study_list_json, 1)

  @parameterized.parameters(
      (vz.StudyState.ABORTED,),
      (vz.StudyState.ACTIVE,),
  )
  def test_set_and_get_study_state(self, state):
    self.client.set_study_state(state)
    self.assertEqual(self.client.get_study_state(), state)

  def test_delete_study(self):
    self.client.delete_study()
    empty_list_json = self.client.list_studies()
    self.assertEmpty(empty_list_json)

  def test_list_trials(self):
    trial_list = self.client.list_trials()
    self.assertLen(trial_list, 1)

  def test_list_optimal_trials(self):
    for i in range(2, 10):
      metric = study_pb2.Measurement.Metric(
          metric_id='example_metric', value=0.2 * i
      )
      completed_trial = study_pb2.Trial(
          name=resources.TrialResource(self.owner_id, self.study_id, i).name,
          id=str(i),
          state=study_pb2.Trial.State.SUCCEEDED,
          final_measurement=study_pb2.Measurement(metrics=[metric]),
      )

      self.servicer.datastore.create_trial(completed_trial)

    trial_list = self.client.list_optimal_trials()
    self.assertLen(trial_list, 1)

  def test_get_trial(self):
    active_trial = self.client.get_trial(trial_id=1)
    self.assertEqual(
        active_trial, vz.TrialConverter.from_proto(self.active_trial)
    )

  @parameterized.named_parameters(
      ('Infeasible', 'infeasible_reason'), ('Complete', None)
  )
  def test_complete_trial(self, infeasibility_reason):
    final_measurement = vz.Measurement(metrics={'metric': vz.Metric(value=0.1)})
    output_trial = self.client.complete_trial(
        trial_id=1,
        final_measurement=final_measurement,
        infeasibility_reason=infeasibility_reason,
    )

    self.assertEqual(output_trial.status, vz.TrialStatus.COMPLETED)
    self.assertEqual(output_trial.infeasibility_reason, infeasibility_reason)

    # See if the rest of the contents were maintained.
    completed_trial = vz.TrialConverter.from_proto(
        self.servicer.datastore.get_trial(self.active_trial.name)
    )
    self.assertEqual(output_trial, completed_trial)

  def test_should_trial_stop(self):
    # Only trial 1 was ACTIVE, so RandomPolicy will signal to stop it.
    should_stop = self.client.should_trial_stop(trial_id=1)
    self.assertTrue(should_stop)
    self.client.stop_trial(trial_id=1)

    # The op will become recycled after the time period and early stopping will
    # be recomputed again. But RandomPolicy will consider the trial non-ACTIVE
    # and simply return False.
    self.local_server.wait_for_early_stop_recycle_period()
    should_stop_again = self.client.should_trial_stop(trial_id=1)
    self.assertFalse(should_stop_again)

  def test_intermediate_measurement(self):
    updated_trial = self.client.report_intermediate_objective_value(
        step=5,
        elapsed_secs=3.0,
        metric_list=[{'example_metric': 5}],
        trial_id=1,
    )
    self.assertLen(updated_trial.measurements, 1)
    self.assertEqual(updated_trial.measurements[0].steps, 5)
    self.assertEqual(updated_trial.measurements[0].elapsed_secs, 3.0)
    self.assertEqual(
        updated_trial.measurements[0].metrics['example_metric'],
        vz.Metric(value=5.0),
    )
    self.assertEqual(updated_trial.id, 1)

  def test_get_suggestions(self):
    suggestion_count = 2
    suggestions_list = self.client.get_suggestions(
        suggestion_count=suggestion_count
    )
    self.assertLen(suggestions_list, suggestion_count)
    logging.info('Suggestions List: %s', suggestions_list)

  # Only test algorithms which don't depend on external libraries (except for
  # numpy).
  @parameterized.parameters(
      dict(algorithm=vz.Algorithm.RANDOM_SEARCH),
      dict(algorithm=vz.Algorithm.QUASI_RANDOM_SEARCH),
      dict(algorithm=vz.Algorithm.GRID_SEARCH),
      dict(algorithm=vz.Algorithm.NSGA2, multiobj=True),
      dict(algorithm=vz.Algorithm.ALGORITHM_UNSPECIFIED, multiobj=True),
      dict(algorithm='DEFAULT', multiobj=True),
      dict(algorithm='DEFAULT', multiobj=False),
      dict(algorithm=vz.Algorithm.GAUSSIAN_PROCESS_BANDIT, multiobj=True),
      dict(algorithm=vz.Algorithm.GAUSSIAN_PROCESS_BANDIT, multiobj=False),
  )
  def test_e2e_tuning(
      self,
      *,
      algorithm,
      num_iterations: int = 10,
      batch_size: int = 1,
      multiobj: bool = False,
  ):
    # Runs end-to-end tuning via back-and-forth communication to server.
    def learning_curve_generator(learning_rate: float) -> List[float]:
      return [learning_rate * step for step in range(10)]

    study_config = vz.StudyConfig()
    study_config.search_space.root.add_float_param(
        'learning_rate', min_value=0.0, max_value=1.0, default_value=0.5
    )
    study_config.search_space.root.add_int_param(
        'num_layers', min_value=1, max_value=5
    )
    study_config.metric_information = [
        vz.MetricInformation(
            name='accuracy', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    ]
    if multiobj:
      study_config.metric_information.append(
          vz.MetricInformation(
              name='latency', goal=vz.ObjectiveMetricGoal.MINIMIZE
          )
      )
    study_config.algorithm = algorithm

    cifar10_client = vizier_client.create_or_load_study(
        owner_id=self.owner_id,
        study_id='cifar10',
        study_config=study_config,
        client_id='cifar10_client',
    )

    for _ in range(num_iterations):
      suggestions = cifar10_client.get_suggestions(suggestion_count=batch_size)
      for trial in suggestions:
        learning_rate = trial.parameters.get_value('learning_rate')
        num_layers = trial.parameters.get_value('num_layers')
        curve = learning_curve_generator(learning_rate)
        for i in range(len(curve)):
          cifar10_client.report_intermediate_objective_value(
              step=i,
              elapsed_secs=0.1 * i,
              metric_list=[{'accuracy': curve[i], 'latency': 0.5 * num_layers}],
              trial_id=trial.id,
          )
        cifar10_client.complete_trial(trial_id=trial.id)

        # Recover the trial from database.
        study_resource = resources.StudyResource.from_name(
            cifar10_client.study_resource_name
        )
        trial_name = resources.TrialResource(
            study_resource.owner_id, study_resource.study_id, trial.id
        ).name
        stored_trial = self.servicer.datastore.get_trial(trial_name)
        stored_curve = [m.metrics[0].value for m in stored_trial.measurements]

        # See if curve was stored correctly.
        self.assertEqual(curve, stored_curve)

        # See if final_measurement is defaulted to end of curve.
        final_accuracy = stored_trial.final_measurement.metrics[0].value
        self.assertEqual(curve[-1], final_accuracy)

  def test_update_metadata(self):
    # Only a smoke test, same as in `service_policy_supporter_test.py`.
    on_study_metadata = vz.Metadata()
    on_study_metadata.ns('bar')['foo'] = '.bar.foo.1'
    on_trial1_metadata = vz.Metadata()
    on_trial1_metadata.ns('bax')['nerf'] = '1.bar.nerf.2'
    metadata_delta = vz.MetadataDelta(
        on_study=on_study_metadata, on_trials={1: on_trial1_metadata}
    )
    self.client.update_metadata(metadata_delta)

  def test_unset_endpoint_client(self):
    study_id = 'dummy_study'
    study_config = vz.StudyConfig()
    study_resource_name = resources.StudyResource(self.owner_id, study_id).name

    vizier_client.environment_variables.server_endpoint = constants.NO_ENDPOINT
    vizier_client.environment_variables.servicer_use_sql_ram()
    # Check if servicer is stored in client.
    local_client1 = vizier_client.create_or_load_study(
        owner_id=self.owner_id,
        client_id='local_client1',
        study_id=study_id,
        study_config=study_config,
    )
    self.assertIsInstance(
        local_client1._service, vizier_service_pb2_grpc.VizierServiceServicer
    )

    # Check if the local server is shared.
    local_client2 = vizier_client.VizierClient(
        study_resource_name=study_resource_name, client_id='local_client2'
    )
    self.assertEqual(local_client1._service, local_client2._service)

    # Same server still exists globally in cache after clients are deleted.
    del local_client1
    del local_client2
    local_client3 = vizier_client.VizierClient(
        study_resource_name=study_resource_name, client_id='local_client3'
    )
    self.assertLen(local_client3.list_studies(), 1)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Comparison test for algorithms using score analysis.

Ex: Typical Efficiency Convergence Test Example
-----------------------------------------------
baseline_factory = benchmarks.BenchmarkStateFactory(...)
candidate_factory = benchmarks.BenchmarkStateFactory(...)

# Run each algorithm for 100 Trials with 5 repeats each.
comparator = comparator_runner.EfficiencyComparisonTester(
        num_trials=100, num_repeats=5)
comparator.assert_better_efficiency(candidate_factory, baseline_factory)

NOTE: assert_converges_faster is a generic method name that conveys the general
use of the class.
"""

from absl import logging
import attr
from jax import random
import numpy as np
from vizier import benchmarks
from vizier import pyvizier as vz
from vizier._src.algorithms.optimizers import vectorized_base as vb
from vizier._src.benchmarks.analyzers import simple_regret_score
from vizier.benchmarks import analyzers
from vizier.pyvizier import converters


class FailedComparisonTestError(Exception):
  """Exception raised for comparison test fails."""


class FailedSimpleRegretConvergenceTestError(Exception):
  """Exception raised for simple-regret convergence test fails."""


@attr.define
class EfficiencyComparisonTester:
  """Comparison test between algorithms using analysis scores."""
  num_trials: int = attr.field(
      default=1, validator=attr.validators.instance_of(int))
  num_repeats: int = attr.field(
      default=1, validator=attr.validators.instance_of(int))

  def assert_better_efficiency(
      self,
      candidate_state_factory: benchmarks.BenchmarkStateFactory,
      baseline_state_factory: benchmarks.BenchmarkStateFactory,
      score_threshold: float = 0.0) -> None:
    """Asserts that candidate is better than baseline via log_eff_score."""
    # TODO: Consider making this more flexible with more runners
    # And enable multimetric.
    runner = benchmarks.BenchmarkRunner(
        benchmark_subroutines=[benchmarks.GenerateAndEvaluate()],
        num_repeats=self.num_trials)

    baseline_curves = []
    candidate_curves = []
    for _ in range(self.num_repeats):
      baseline_state = baseline_state_factory()
      candidate_state = candidate_state_factory()

      baseline_statement = baseline_state.experimenter.problem_statement()
      if len(baseline_statement.metric_information) > 1:
        raise ValueError('Support for multimetric is not yet')
      if baseline_statement != (
          candidate_statement :=
          candidate_state.experimenter.problem_statement()):
        raise ValueError('Comparison tests done for different statements: '
                         f'{baseline_statement} vs {candidate_statement}')

      runner.run(baseline_state)
      runner.run(candidate_state)
      baseline_curves.append(
          analyzers.ConvergenceCurveConverter(
              baseline_statement.metric_information.item()
          ).convert(baseline_state.algorithm.supporter.GetTrials())
      )
      candidate_curves.append(
          analyzers.ConvergenceCurveConverter(
              baseline_statement.metric_information.item()
          ).convert(candidate_state.algorithm.supporter.GetTrials())
      )

    baseline_curve = analyzers.ConvergenceCurve.align_xs(
        baseline_curves, interpolate_repeats=True
    )[0]
    candidate_curve = analyzers.ConvergenceCurve.align_xs(
        candidate_curves, interpolate_repeats=True
    )[0]
    comparator = analyzers.LogEfficiencyConvergenceCurveComparator(
        baseline_curve=baseline_curve, compared_curve=candidate_curve
    )

    if (log_eff_score := comparator.score()) < score_threshold:
      raise FailedComparisonTestError(
          f'Log efficiency score {log_eff_score} is less than {score_threshold}'
          f' when comparing algorithms: {candidate_state_factory} '
          f'vs baseline of {baseline_state_factory} for {self.num_trials} '
          f' Trials with {self.num_repeats} repeats')


@attr.define(kw_only=True)
class SimpleRegretComparisonTester:
  """Compare two algorithms by their simple regrets.

  The test runs the baseline algorithm 'baseline_num_repeats' times each with
  'baseline_num_trials' trials and computes the simple regret in each trial,
  and similarly for the candidate algorithm.

  A one-sided T-test is performed to compute the p-value of observing the
  difference in the simple regret sample means. The T-test score (p-value) is
  compared against the significance level (alpha) to determine if the test
  passed.
  """
  baseline_num_trials: int
  candidate_num_trials: int
  baseline_suggestion_batch_size: int
  candidate_suggestion_batch_size: int
  baseline_num_repeats: int
  candidate_num_repeats: int
  alpha: float = attr.field(
      validator=attr.validators.and_(
          attr.validators.ge(0), attr.validators.le(0.1)),
      default=0.05)
  goal: vz.ObjectiveMetricGoal

  def assert_optimizer_better_simple_regret(
      self,
      converter: converters.TrialToModelInputConverter,
      score_fn: vb.ArrayScoreFunction,
      baseline_strategy_factory: vb.VectorizedStrategyFactory,
      candidate_strategy_factory: vb.VectorizedStrategyFactory,
  ) -> None:
    """Assert if candidate optimizer has better simple regret than the baseline.
    """
    baseline_obj_values = []
    candidate_obj_values = []

    baseline_optimizer_factory = vb.VectorizedOptimizerFactory(
        baseline_strategy_factory,
        suggestion_batch_size=self.baseline_suggestion_batch_size,
        max_evaluations=self.baseline_num_trials,
    )
    candidate_optimizer_factory = vb.VectorizedOptimizerFactory(
        candidate_strategy_factory,
        suggestion_batch_size=self.candidate_suggestion_batch_size,
        max_evaluations=self.candidate_num_trials,
    )
    baseline_optimizer = baseline_optimizer_factory(converter)
    candidate_optimizer = candidate_optimizer_factory(converter)

    for i in range(self.baseline_num_repeats):
      res = baseline_optimizer(score_fn, count=1, seed=random.PRNGKey(i))  # pytype: disable=wrong-arg-types
      trial = vb.best_candidates_to_trials(res, converter)
      baseline_obj_values.append(
          trial[0].final_measurement.metrics['acquisition'].value)

    for i in range(self.candidate_num_repeats):
      res = candidate_optimizer(score_fn, count=1, seed=random.PRNGKey(i))  # pytype: disable=wrong-arg-types
      trial = vb.best_candidates_to_trials(res, converter)
      candidate_obj_values.append(
          trial[0].final_measurement.metrics['acquisition'].value)

    self._conclude_test(baseline_obj_values, candidate_obj_values)

  def assert_benchmark_state_better_simple_regret(
      self,
      baseline_benchmark_state_factory: benchmarks.BenchmarkStateFactory,
      candidate_benchmark_state_factory: benchmarks.BenchmarkStateFactory,
  ) -> None:
    """Runs simple-regret convergence test for benchmark state."""

    def _run_one(benchmark_state_factory: benchmarks.BenchmarkStateFactory,
                 num_trials: int, batch_size: int, seed: int) -> float:
      """Run one benchmark run and returns simple regret."""
      benchmark_state = benchmark_state_factory(seed=seed)
      baseline_runner = benchmarks.BenchmarkRunner(
          benchmark_subroutines=[benchmarks.GenerateAndEvaluate(batch_size)],
          num_repeats=num_trials // batch_size)
      baseline_runner.run(benchmark_state)
      # Extract best metric
      best_trial = benchmark_state.algorithm.supporter.GetBestTrials(count=1)[0]
      metric_name = benchmark_state.experimenter.problem_statement(
      ).single_objective_metric_name
      return best_trial.final_measurement.metrics[metric_name].value

    baseline_obj_values = []
    candidate_obj_values = []

    for idx in range(self.baseline_num_repeats):
      baseline_obj_values.append(
          _run_one(
              benchmark_state_factory=baseline_benchmark_state_factory,
              num_trials=self.baseline_num_trials,
              batch_size=self.baseline_suggestion_batch_size,
              seed=idx))

    for idx in range(self.candidate_num_repeats):
      candidate_obj_values.append(
          _run_one(
              benchmark_state_factory=candidate_benchmark_state_factory,
              num_trials=self.candidate_num_trials,
              batch_size=self.candidate_suggestion_batch_size,
              seed=idx))
    self._conclude_test(baseline_obj_values, candidate_obj_values)

  def _conclude_test(self, baseline_obj_values: list[float],
                     candidate_obj_values: list[float]) -> None:
    """Concludes test based on baseline and candidate objective func values."""

    p_value = simple_regret_score.t_test_mean_score(baseline_obj_values,
                                                    candidate_obj_values,
                                                    self.goal)
    msg = self._generate_summary(baseline_obj_values, candidate_obj_values,
                                 p_value)
    if p_value <= self.alpha:
      logging.info('Convergence test PASSED:\n %s', msg)
    else:
      raise FailedSimpleRegretConvergenceTestError(msg)

  def _generate_summary(
      self,
      baseline_obj_values: list[float],
      candidate_obj_values: list[float],
      p_value: float,
  ) -> str:
    """Generate summary message."""
    baseline_mean = np.mean(baseline_obj_values)
    baseline_std = np.std(baseline_obj_values)
    candidate_mean = np.mean(candidate_obj_values)
    candidate_std = np.std(candidate_obj_values)
    return (f'\nObjective goal={self.goal.name}'
            f'\nP-value={p_value}'
            f'\nAlpha={self.alpha}'
            f'\nBaseline Objective Mean: {baseline_mean}'
            f'\nBaseline Objective Std: {baseline_std}'
            f'\nCandidate Objective Mean: {candidate_mean}'
            f'\nCandidate Objective Std: {candidate_std}'
            f'\nBaseline Objective Scores: {baseline_obj_values}'
            f'\nCandidate Objective Scores: {candidate_obj_values}')
